"""This module is used to control Vici Valco Universal Electronic Actuators."""
from __future__ import annotations

from dataclasses import dataclass

import aioserial
from loguru import logger

from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.vicivalco.vici_valve_component import ViciInjectionValve
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import dario, jakob, wei_hsin


@dataclass
class ViciCommand:
    """This class represent a command. Its bytes() method is transmitted to the valve."""

    command: str
    valve_id: int | None = None
    value: str = ""
    reply_lines: int = 1

    def __str__(self) -> str:
        """Provide a string representation of the command used, nice for logs."""
        address = str(self.valve_id) if self.valve_id is not None else ""
        return f"{address} {self.command}{self.value}"

    def __bytes__(self) -> bytes:
        """Byte representation of the command used for serial communication."""
        return str(self).encode("ascii")


class ViciValcoValveIO:
    """Setup with serial parameters, low level IO."""

    DEFAULT_CONFIG = {
        "timeout": 0.5,
        "baudrate": 9600,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }

    def __init__(self, aio_port: aioserial.Serial) -> None:
        """Initialize communication on the serial port where the valves are located and initialize them.

        Args:
        ----
            aio_port: aioserial.Serial() object
        """
        self._serial = aio_port

    @classmethod
    def from_config(cls, port, **serial_kwargs):
        """Create ViciValcoValveIO from config."""
        # Merge default serial settings with provided ones.
        configuration = dict(ViciValcoValveIO.DEFAULT_CONFIG, **serial_kwargs)

        try:
            serial_object = aioserial.AioSerial(port, **configuration)
        except aioserial.SerialException as serial_exception:
            raise InvalidConfigurationError(
                f"Could not open serial port {port} with configuration {configuration}"
            ) from serial_exception

        return cls(serial_object)

    async def _read_reply(self, lines: int) -> str:
        """Read the valve reply from serial communication."""
        reply_string = ""
        for _ in range(lines):
            line = await self._serial.readline_async()
            reply_string += line.decode("ascii")

        if reply_string:
            logger.debug(f"Reply received: {reply_string}")
        else:
            raise InvalidConfigurationError(
                "No response received from valve! Check valve address?"
            )

        return reply_string.rstrip()

    async def write_and_read_reply(self, command: ViciCommand) -> str:
        """Write command to valve and read reply."""
        # Make sure input buffer is empty
        self._serial.reset_input_buffer()

        # Send command
        await self._serial.write_async(bytes(command))
        logger.debug(f"Command {command} sent!")

        if command.reply_lines == 0:
            return ""
        else:
            return await self._read_reply(command.reply_lines)

    @property
    def name(self) -> str:
        """Provide a nice-looking default name to valve based on its serial connection."""
        try:
            return self._serial.name
        except AttributeError:
            return ""


class ViciValve(FlowchemDevice):
    """ViciValco injection valves."""

    # This class variable is used for daisy chains (i.e. multiple valves on the same serial connection). Details below.
    _io_instances: set[ViciValcoValveIO] = set()
    # When several valves are daisy-chained on the same serial port, they need to all access the *same* Serial object,
    # because access to the serial port is exclusive by definition.
    # The mutable object _io_instances as class variable creates a shared state across all the instances.

    def __init__(
        self,
        valve_io: ViciValcoValveIO,
        name: str = "",
        address: int | None = None,
    ) -> None:
        """Create instance from an existing ViciValcoValveIO object. This allows dependency injection.

        See from_config() class method for config-based init.

        Args:
        ----
            valve_io: An ViciValcoValveIO w/ serial connection to the daisy chain w/ target valve.
            address: number of valve in array, 1 for first one, auto-assigned on init based on position.
            name: 'cause naming stuff is important.
        """
        self.valve_io = valve_io
        ViciValve._io_instances.add(self.valve_io)

        # The valve name is used for logs and error messages.
        self.name = name if name else f"Valve {self.valve_io.name}:{address}"
        super().__init__(name=name)  # type: ignore

        self.address = address
        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Vici-Valco",
            model="Universal Valve Actuator",
        )

    @classmethod
    def from_config(
        cls,
        port: str,
        address: int,
        name: str = "",
        **serial_kwargs,
    ):
        """Create instances via provided parameters to enable programmatic instantiation."""
        existing_io = [v for v in ViciValve._io_instances if v._serial.port == port]

        # If no existing serial object are available for the port provided, create a new one
        valve_io = (
            existing_io.pop()
            if existing_io
            else ViciValcoValveIO.from_config(port, **serial_kwargs)
        )

        return cls(valve_io, address=address, name=name)

    async def initialize(self):
        """Must be called after init before anything else."""
        # Learning positions is only needed if the valve head has been reinstalled.
        await self.learn_positions()

        # Homing implies moving to position 1.
        await self.home()

        # Test connectivity by querying the valve's firmware version
        self.device_info.version = await self.version()
        logger.info(f"Connected to {self.name} - FW ver.: {self.device_info.version}!")

        # Add component
        self.components.append(ViciInjectionValve("injection-valve", self))

    async def learn_positions(self) -> None:
        """Initialize valve only, there is no reply -> reply_lines = 0."""
        learn = ViciCommand(valve_id=self.address, command="LRN")
        await self.valve_io.write_and_read_reply(learn)

    async def home(self) -> None:
        """Initialize valve only: Move to Home position."""
        home = ViciCommand(valve_id=self.address, command="HM")
        await self.valve_io.write_and_read_reply(home)

        # This seems necessary to make sure move is finished
        await self.get_raw_position()

    async def version(self) -> str:
        """Return the current firmware version reported by the valve."""
        version = ViciCommand(valve_id=self.address, command="VR", reply_lines=5)
        return await self.valve_io.write_and_read_reply(version)

    async def get_raw_position(self) -> str:
        """Represent the position of the valve."""
        current_pos = ViciCommand(valve_id=self.address, command="CP")
        return await self.valve_io.write_and_read_reply(current_pos)

    async def set_raw_position(self, position: str):
        """Set valve position."""
        valve_by_name_cw = ViciCommand(
            valve_id=self.address,
            command="GO",
            value=position,
            reply_lines=0,
        )
        await self.valve_io.write_and_read_reply(valve_by_name_cw)

    async def timed_toggle(self, injection_time: str):
        """Switch valve to a position for a given time."""
        delay = ureg.Quantity(injection_time).to("ms")
        set_delay = ViciCommand(
            valve_id=self.address,
            command="DT",
            value=delay.magnitude,
        )
        await self.valve_io.write_and_read_reply(set_delay)

        time_toggle = ViciCommand(valve_id=self.address, command="TT")
        await self.valve_io.write_and_read_reply(time_toggle)


if __name__ == "__main__":
    import asyncio

    valve1 = ViciValve.from_config(port="COM13", address=0, name="test1")
    asyncio.run(valve1.initialize())

    # Set position works with both strings and InjectionValvePosition
    asyncio.run(valve1.set_raw_position("2"))



"""Pressure sensor."""
from flowchem.devices.flowchem_device import FlowchemDevice

from .sensor import Sensor


class PressureSensor(Sensor):
    """
    A class to represent a pressure sensor.

    Attributes:
    -----------
    hw_device : FlowchemDevice
        The hardware device this sensor is interfacing with.

    Methods:
    --------
    read_pressure(units: str = "bar"):
        Read the current pressure from the sensor and return it in the specified units.
    """

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Constructs all the necessary attributes for the PressureSensor object.

        Parameters:
        -----------
        name : str
            The name of the pressure sensor.
        hw_device : FlowchemDevice
            The hardware device this sensor is interfacing with.
        """
        super().__init__(name, hw_device)
        self.add_api_route("/read-pressure", self.read_pressure, methods=["GET"])

        # Ontology: Pressure Sensor Device
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/NCIT_C50167",
        )

    async def read_pressure(self, units: str = "bar"):
        """
        Read the current pressure from the sensor and return it in the specified units.

        Parameters:
        -----------
        units : str, optional
            The units in which to return the pressure (default is bar).

        Returns:
        --------
        float
            The current pressure in the specified units.
        """
        ...



"""Light control."""
from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.flowchem_component import FlowchemComponent

if TYPE_CHECKING:
    from flowchem.devices.flowchem_device import FlowchemDevice


class Photoreactor(FlowchemComponent):
    """A generic photoreactor."""

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(name, hw_device)

        self.add_api_route("/intensity", self.set_intensity, methods=["PUT"])
        self.add_api_route("/intensity", self.get_intensity, methods=["GET"])

        self.add_api_route("/power-on", self.power_on, methods=["PUT"])
        self.add_api_route("/power-off", self.power_off, methods=["PUT"])

    async def set_intensity(self, percent: int):
        """Set light intensity (in percent)."""
        ...

    async def get_intensity(self) -> int:  # type: ignore
        """Get current light intensity (in percent)."""
        ...

    async def power_on(self):
        """Turn on light."""
        ...

    async def power_off(self):
        """Turn off light."""
        ...



"""This module is used to control Harvard Apparatus Elite 11 syringe pump via the 11 protocol."""
from __future__ import annotations

import asyncio
import warnings

import pint
from loguru import logger
from pydantic import BaseModel

from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.harvardapparatus._pumpio import (
    HarvardApparatusPumpIO,
    Protocol11Command,
    PumpStatus,
)
from flowchem.devices.harvardapparatus.elite11_pump import (
    Elite11PumpOnly,
    Elite11PumpWithdraw,
)
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import dario, jakob, wei_hsin


class PumpInfo(BaseModel):
    """Detailed pump info. e.g.:

    ('Pump type          Pump 11',
    'Pump type string   11 ELITE I/W Single',
    'Display type       Sharp',
    'Steps per rev      400',
    'Gear ratio         1:1',
    'Pulley ratio       2.4:1',
    'Lead screw         24 threads per inch',
    'Microstepping      16 microsteps per step',
    'Low speed limit    27 seconds',
    'High speed limit   26 microseconds',
    'Motor polarity     Reverse',
    'Min syringe size   0.1 mm',
    'Max syringe size   33 mm',
    'Min raw force %    20%',
    'Max raw force %    80%',
    'Encoder            100 lines',
    'Direction          Infuse/withdraw',
    'Programmable       Yes',
    'Limit switches     No',
    'Command set        None', '')
    """

    pump_type: str
    pump_description: str
    infuse_only: bool

    @classmethod
    def parse_pump_string(cls, metrics_text: list[str]):
        """Parse pump response string into model."""
        pump_type, pump_description, infuse_only = "", "", True
        for line in metrics_text:
            if line.startswith("Pump type  "):
                pump_type = line[9:].strip()
            elif line.startswith("Pump type string"):
                pump_description = line[16:].strip()
            elif line.startswith("Direction"):
                infuse_only = "withdraw" not in line
        return cls(
            pump_type=pump_type,
            pump_description=pump_description,
            infuse_only=infuse_only,
        )


class Elite11(FlowchemDevice):
    """Controls Harvard Apparatus Elite11 syringe pumps.

    The same protocol (Protocol11) can be used on other HA pumps, but is untested.
    Several pumps can be daisy-chained on the same serial connection, if so address 0 must be the first one.
    Read the manufacturer manual for more details.
    """

    # This class variable is used for daisy chains (i.e. multiple pumps on the same serial connection).
    _io_instances: set[HarvardApparatusPumpIO] = set()

    def __init__(
        self,
        pump_io: HarvardApparatusPumpIO,
        syringe_diameter: str = "",
        syringe_volume: str = "",
        address: int = 0,
        name: str = "",
        force: int = 30,
    ) -> None:
        super().__init__(name)

        # Create communication
        self.pump_io = pump_io
        Elite11._io_instances.add(self.pump_io)

        self.address = address
        self._infuse_only = False  # Actual value set in initialize

        # syringe diameter and volume, and force will be set in initialize()
        self._force = force
        if syringe_diameter:
            self._diameter = syringe_diameter
        else:
            raise InvalidConfigurationError("Please provide the syringe diameter!")

        if syringe_volume:
            self._syringe_volume = syringe_volume
        else:
            raise InvalidConfigurationError("Please provide the syringe volume!")

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="HarvardApparatus",
            model="Elite11",
        )

    @classmethod
    def from_config(
        cls,
        port: str,
        syringe_diameter: str,
        syringe_volume: str,
        address: int = -1,
        name: str = "",
        force: int = 30,
        **serial_kwargs,
    ):
        """Programmatic instantiation from configuration.

        Many pump can be present on the same serial port with different addresses.
        This shared list of PumpIO objects allow shared state in a borg-inspired way, avoiding singletons
        This is only relevant to programmatic instantiation, i.e. when from_config() is called per each pump from a
        config file, as it is the case in the HTTP server.
        Pump_IO() manually instantiated are not accounted for.
        """
        pumpio = None
        for obj in Elite11._io_instances:
            if obj._serial.port == port:
                pumpio = obj
                break

        # If not existing serial object are available for the port provided, create a new one
        if pumpio is None:
            pumpio = HarvardApparatusPumpIO(port, **serial_kwargs)

        return cls(
            pumpio,
            address=address,
            name=name,
            syringe_diameter=syringe_diameter,
            syringe_volume=syringe_volume,
            force=force,
        )

    async def initialize(self):
        """Initialize Elite11.

        Query model and version number of firmware to check if pump is connected.
        Responds with a load of stuff, but the last three characters
        are the prompt XXY, where XX is the address and Y is pump status.
        The status can be one of the three: [":", ">" "<"] respectively
        when stopped, running forwards (pumping), or backwards (withdrawing).
        The prompt is used to confirm that the address is correct.
        """
        # Autodetect address if none provided
        if self.address == -1:
            self.address = self.pump_io.autodiscover_address()

        # Test communication and return InvalidConfiguration on failure
        try:
            await self.stop()
        except IndexError as index_e:
            raise InvalidConfigurationError(
                f"Check pump address! Currently {self.address=}"
            ) from index_e

        # Sets syringe parameters
        await self.set_syringe_diameter(ureg.Quantity(self._diameter))
        await self.set_syringe_volume(ureg.Quantity(self._syringe_volume))
        await self.set_force(self._force)

        logger.info(
            f"Connected to '{self.name}'! [{self.pump_io._serial.name}:{self.address}]",
        )
        version = await self.version()
        self._infuse_only = "I/W" not in version

        # Clear target volume eventually set to prevent pump from stopping prematurely
        await self.set_target_volume("0 ml")

        # Add components
        if self._infuse_only:
            self.components.append(Elite11PumpOnly("pump", self))
        else:
            self.components.append(Elite11PumpWithdraw("pump", self))

    @staticmethod
    def _parse_version(version_text: str) -> tuple[int, int, int]:
        """Extract semver from Elite11 version string, e.g. '11 ELITE I/W Single 3.0.4'."""
        version = version_text.split(" ")[-1]
        digits = version.split(".")
        return int(digits[0]), int(digits[1]), int(digits[2])

    async def _send_command_and_read_reply(
        self,
        command: str,
        parameter="",
        parse=True,
        multiline=False,
    ):
        """Send a command based on its template and return the corresponding reply as str."""
        cmd = Protocol11Command(
            command=command,
            pump_address=self.address,
            arguments=parameter,
        )
        reply = await self.pump_io.write_and_read_reply(cmd, return_parsed=parse)
        if multiline:
            return reply
        else:
            return reply[0]

    async def get_syringe_diameter(self) -> str:
        """Get syringe diameter in mm. A value between 1 and 33 mm."""
        return await self._send_command_and_read_reply("diameter")

    async def set_syringe_diameter(self, diameter: pint.Quantity):
        """Set syringe diameter. This can be set in the interval 1 mm to 33 mm."""
        if not 1 * ureg.mm <= diameter <= 33 * ureg.mm:
            logger.warning(
                f"Invalid diameter provided: {diameter}! [Valid range: 1-33 mm]",
            )
            return False

        await self._send_command_and_read_reply(
            "diameter",
            parameter=f"{diameter.to('mm').magnitude:.4f} mm",
        )
        return None

    async def get_syringe_volume(self) -> str:
        """Return the syringe volume as str w/ units."""
        return await self._send_command_and_read_reply("svolume")  # e.g. '100 ml'

    async def set_syringe_volume(self, volume: pint.Quantity):
        """Set the syringe volume in ml."""
        await self._send_command_and_read_reply(
            "svolume",
            parameter=f"{volume.m_as('ml'):.15f} m",
        )

    async def get_force(self):
        """Pump force, in percentage.

        Manufacturer suggested values are:
            stainless steel:    100%
            plastic syringes:   50% if volume <= 5 ml else 100%
            glass/glass:        30% if volume <= 20 ml else 50%
            glass/plastic:      30% if volume <= 250 ul, 50% if volume <= 5ml else 100%
        """
        percent = await self._send_command_and_read_reply("FORCE")
        return int(percent[:-1])

    async def set_force(self, force_percent: int):
        """Set the pump force, see `Elite11.get_force()` for suggested values."""
        await self._send_command_and_read_reply(
            "FORCE",
            parameter=str(int(force_percent)),
        )

    async def _bound_rate_to_pump_limits(self, rate: str) -> float:
        """Bound the rate provided to pump's limit.

        These are function of the syringe diameter.
        NOTE: Infusion and withdraw limits are equal!
        """
        # Get current pump limits (those are function of the syringe diameter)
        limits_raw = await self._send_command_and_read_reply("irate lim")

        # Lower limit usually expressed in nl/min so unit-aware quantities are needed
        lower_limit, upper_limit = map(ureg, limits_raw.split(" to "))

        # add units by assumption
        # if rate.isnumeric():
        #     rate = flowrate + "ml/min"
        #     logger.warning("No units provided, assuming milliliter/minutes.")
        # parsed_f = ureg.Quantity(flow)

        # Also add units to the provided rate
        set_rate = ureg.Quantity(rate)

        # Bound rate to acceptance range
        if set_rate < lower_limit:
            logger.warning(
                f"The requested rate {rate} is lower than the minimum possible ({lower_limit})!"
                f"Setting rate to {lower_limit} instead!",
            )
            set_rate = lower_limit

        if set_rate > upper_limit:
            logger.warning(
                f"The requested rate {rate} is higher than the maximum possible ({upper_limit})!"
                f"Setting rate to {upper_limit} instead!",
            )
            set_rate = upper_limit

        return set_rate.to("ml/min").magnitude

    async def version(self) -> str:
        """Return the current firmware version reported by the pump."""
        return await self._send_command_and_read_reply(
            "VER",
        )  # '11 ELITE I/W Single 3.0.4

    async def is_moving(self) -> bool:
        """Evaluate prompt for current status, i.e. moving or not."""
        status = await self._send_command_and_read_reply(" ", parse=False)
        prompt = PumpStatus(status[2:3])
        return prompt in (PumpStatus.INFUSING, PumpStatus.WITHDRAWING)

    async def infuse(self):
        """Run pump in infuse mode."""
        await self._send_command_and_read_reply("irun")
        logger.info("Pump infusion started!")
        return True

    async def withdraw(self):
        """Activate pump to run in withdraw mode."""
        await self._send_command_and_read_reply("wrun")
        logger.info("Pump withdraw started!")
        return True

    async def stop(self):
        """Stop pump."""
        await self._send_command_and_read_reply("stp")
        logger.info("Pump stopped")

    async def wait_until_idle(self):
        """Wait until the pump is not moving."""
        while await self.is_moving():
            await asyncio.sleep(0.05)

    async def get_flow_rate(self) -> float:
        """Return the infusion rate as str w/ units."""
        flow_value = await self._send_command_and_read_reply("irate")
        flowrate = ureg.Quantity(flow_value)
        logger.debug(f"Current infusion flow rate is {flowrate}")
        return flowrate.m_as("ml/min")

    async def set_flow_rate(self, rate: str):
        """Set the infusion rate."""
        set_rate = await self._bound_rate_to_pump_limits(rate=rate)
        await self._send_command_and_read_reply(
            "irate",
            parameter=f"{set_rate:.10f} m/m",
        )

    async def get_withdrawing_flow_rate(self) -> float:
        """Return the withdrawing flow rate as ml/min."""
        flow_value = await self._send_command_and_read_reply("wrate")
        flowrate = ureg.Quantity(flow_value)
        logger.debug(f"Current withdraw flow rate is {flowrate}")
        return flowrate.m_as("ml/min")

    async def set_withdrawing_flow_rate(self, rate: str):
        """Set the infusion rate."""
        set_rate = await self._bound_rate_to_pump_limits(rate=rate)
        await self._send_command_and_read_reply("wrate", parameter=f"{set_rate} m/m")

    async def set_target_volume(self, volume: str):
        """Set target volume in ml. If the volume is set to 0, the target is cleared."""
        await self._send_command_and_read_reply("cvolume")

        target_volume = ureg.Quantity(volume)
        if target_volume.magnitude == 0:
            await self._send_command_and_read_reply("ctvolume")
        else:
            set_vol = await self._send_command_and_read_reply(
                "tvolume",
                parameter=f"{target_volume.m_as('ml')} m",
            )
            if "Argument error" in set_vol:
                warnings.warn(
                    f"Cannot set target volume of {target_volume} with a "
                    f"{self.get_syringe_volume()} syringe!",
                    stacklevel=2,
                )

    async def pump_info(self) -> PumpInfo:
        """Return pump info."""
        parsed_multiline_response = await self._send_command_and_read_reply(
            "metrics",
            multiline=True,
        )
        return PumpInfo.parse_pump_string(parsed_multiline_response)


if __name__ == "__main__":
    pump = Elite11.from_config(
        port="COM5",
        syringe_volume="10 ml",
        syringe_diameter="14.567 mm",
        address=3,
    )

    async def main():
        """Test function."""
        await pump.initialize()
        # await pump.set_syringe_diameter("30 mm")
        # await pump.set_syringe_diameter("30 mm")
        print(await pump.version())
        await pump.set_flow_rate("0.001 ml/min")
        await pump.set_target_volume("0.001 ml")
        await pump.infuse()
        await asyncio.sleep(2)
        await pump.pump_info()

    asyncio.run(main())



"""Zeroconf (mDNS) server."""
import uuid

from loguru import logger
from zeroconf import (
    IPVersion,
    NonUniqueNameException,
    ServiceInfo,
    Zeroconf,
    get_all_addresses,
)


class ZeroconfServer:
    """Server to advertise Flowchem devices via zero configuration networking."""

    def __init__(self, port: int = 8000) -> None:
        # Server properties
        self.port = port
        self.server = Zeroconf(ip_version=IPVersion.V4Only)

        # Get list of host addresses
        self.mdns_addresses = [
            ip
            for ip in get_all_addresses()  # Get all local IP
            if ip not in ("127.0.0.1", "0.0.0.0")
            and not ip.startswith("169.254")  # Remove invalid IPs
        ]
        if not self.mdns_addresses:
            self.mdns_addresses.append("127.0.0.1")

        logger.info(f"Zeroconf server up, broadcasting on IPs: {self.mdns_addresses}")

    async def add_device(self, name: str) -> None:
        """Add device to the server."""
        properties = {
            "path": rf"http://{self.mdns_addresses[0]}:{self.port}/{name}/",
            "id": f"{name}:{uuid.uuid4()}".replace(" ", ""),
        }

        # LabThing service
        service_info = ServiceInfo(
            type_="_labthing._tcp.local.",
            name=name + "._labthing._tcp.local.",
            port=self.port,
            properties=properties,
            parsed_addresses=self.mdns_addresses,
        )

        try:
            await self.server.async_register_service(service_info)
        except NonUniqueNameException as name_error:
            msg = (
                f"Cannot initialize zeroconf service for '{name}'"
                f"The same name is already in use: you cannot run flowchem twice for the same device!"
            )
            raise RuntimeError(msg) from name_error
        logger.debug(f"Device {name} registered as Zeroconf service!")


if __name__ == "__main__":
    test = ZeroconfServer()
    print("Press Enter to exit.")
    input()



import asyncio
from dataclasses import dataclass
from enum import Enum

import aioserial
from loguru import logger

from flowchem.utils.exceptions import DeviceError, InvalidConfigurationError


class PumpStatus(Enum):
    """Possible pump statuses, as defined by the reply prompt."""

    IDLE = ":"
    INFUSING = ">"
    WITHDRAWING = "<"
    TARGET_REACHED = "T"
    STALLED = "*"


@dataclass
class Protocol11Command:
    """Class representing a pump command."""

    command: str
    pump_address: int
    arguments: str


class HarvardApparatusPumpIO:
    """Setup with serial parameters, low level IO."""

    DEFAULT_CONFIG = {"timeout": 0.1, "baudrate": 115200}

    def __init__(self, port: str, **kwargs) -> None:
        # Merge default settings, including serial, with provided ones.
        configuration = dict(HarvardApparatusPumpIO.DEFAULT_CONFIG, **kwargs)

        self.lock = asyncio.Lock()

        try:
            self._serial = aioserial.AioSerial(port, **configuration)
        except aioserial.SerialException as serial_exception:
            logger.error(f"Cannot connect to the Pump on the port <{port}> issue:{serial_exception}")
            raise InvalidConfigurationError(
                f"Cannot connect to the Pump on the port <{port}>"
            ) from serial_exception

    async def _write(self, command: Protocol11Command):
        """Write a command to the pump."""
        command_msg = f"{command.pump_address}{command.command} {command.arguments}\r\n"

        try:
            await self._serial.write_async(command_msg.encode("ascii"))
        except aioserial.SerialException as serial_exception:
            raise InvalidConfigurationError from serial_exception
        logger.debug(f"Sent {command_msg!r}!")

    async def _read_reply(self) -> list[str]:
        """Read the pump reply from serial communication."""
        reply_string = []

        for line in await self._serial.readlines_async():
            reply_string.append(line.decode("ascii").strip())
            logger.debug(f"Received {line!r}!")

        # First line is usually empty, but some prompts such as T* actually leak into this line sometimes.
        reply_string.pop(0)
        return [x for x in reply_string if x]  # remove empty strings from reply_string

    @staticmethod
    def parse_response_line(line: str) -> tuple[int, PumpStatus, str]:
        """Split a received line in its components: address, prompt and reply body."""
        assert len(line) >= 3
        pump_address, status = int(line[0:2]), PumpStatus(line[2:3])

        # Target reached is the only two-character status
        if status is PumpStatus.TARGET_REACHED:
            return pump_address, status, line[4:]
        return pump_address, status, line[3:]

    @staticmethod
    def parse_response(
        response: list[str],
    ) -> tuple[list[int], list[PumpStatus], list[str]]:
        """Aggregate address prompt and reply body from all the reply lines and return them."""
        parsed_lines = list(map(HarvardApparatusPumpIO.parse_response_line, response))
        return zip(*parsed_lines, strict=True)  # type: ignore

    @staticmethod
    def check_for_errors(response_line, command_sent):
        """Further response parsing, checks for error messages."""
        error_string = (
            "Command error",
            "Unknown command",
            "Argument error",
            "Out of range",
        )
        if any(e in response_line for e in error_string):
            logger.error(
                f"Error for command {command_sent} on pump {command_sent.pump_address}!"
                f"Reply: {response_line}",
            )
            raise DeviceError("Command error")

    async def write_and_read_reply(
        self,
        command: Protocol11Command,
        return_parsed: bool = True,
    ) -> list[str]:
        """Send a command to the pump, read the replies and return it, optionally parsed.

        If unparsed reply is a List[str] with raw replies.
        If parsed reply is a List[str] w/ reply body (address and prompt removed from each line).
        """
        async with self.lock:
            self._serial.reset_input_buffer()
            await self._write(command)
            response = await self._read_reply()

        if not response:
            logger.error("No reply received from pump!")
            raise InvalidConfigurationError(
                "No response received. Is the address right?"
            )

        pump_address, status, parsed_response = self.parse_response(response)

        # All the replies came from the target pump
        assert all(address == command.pump_address for address in pump_address)

        # No stall reply is present
        if PumpStatus.STALLED in status:
            logger.error("Pump stalled!")
            raise DeviceError("Pump stalled! Press display on pump to clear error :(")

        # Check for error in the last response line
        self.check_for_errors(response_line=response[-1], command_sent=command)
        return parsed_response if return_parsed else response

    def autodiscover_address(self) -> int:
        """Autodiscover pump address based on response received."""
        self._serial.write(b"\r\n")
        self._serial.readline()
        prompt = self._serial.readline()
        valid_status = [status.value for status in PumpStatus]
        address = 0 if prompt[0:2].decode() in valid_status else int(prompt[0:2])
        logger.debug(f"Address detected as {address}")
        return address



"""Elite11 pump component."""
from __future__ import annotations

from typing import TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from .elite11 import Elite11
from flowchem.components.pumps.syringe_pump import SyringePump


class Elite11PumpOnly(SyringePump):
    """
    Elite11 pump component without withdraw option.

    Attributes:
        hw_device (Elite11): The hardware device controlling the pump.
    """
    hw_device: Elite11  # for typing's sake

    @staticmethod
    def is_withdrawing_capable():
        """
        Check if the pump is capable of withdrawing.

        Returns:
            bool: False, as this pump is not capable of withdrawing.
        """
        return False

    async def is_pumping(self) -> bool:
        """
        Check if the pump is currently moving.

        Returns:
            bool: True if the pump is moving, False otherwise.
        """
        return await self.hw_device.is_moving()

    async def stop(self):
        """Stop pump."""
        await self.hw_device.stop()

    async def infuse(self, rate: str = "", volume: str = "0 ml") -> bool:
        """
        Infuse at the specified rate and volume.

        Args:
            rate (str): The flow rate for infusion. If not specified, the previous rate will be used.
            volume (str): The target volume for infusion. Defaults to "0 ml".

        Returns:
            bool: True if infusion starts successfully, False otherwise.
        """
        if await self.is_pumping():
            logger.warning("Pump already moving! change to different flow rate!!!")

        if rate:  # Else previous rate will be used
            await self.hw_device.set_flow_rate(rate)

        if volume:
            await self.hw_device.set_target_volume(volume)

        return await self.hw_device.infuse()


class Elite11PumpWithdraw(Elite11PumpOnly):
    """
    Elite11 pump component with withdraw option.

    Attributes:
        hw_device (Elite11): The hardware device controlling the pump.
    """
    @staticmethod
    def is_withdrawing_capable():
        """
        Check if the pump is capable of withdrawing.

        Returns:
            bool: True, as this pump is capable of withdrawing.
        """
        return True

    async def withdraw(self, rate: str = "1 ml/min", volume: str | None = None) -> bool:
        """
        Withdraw at the specified rate and volume.

        Args:
            rate (str): The flow rate for withdrawing. Defaults to "1 ml/min".
            volume (str | None): The target volume for withdrawing. If not specified, the previous volume will be used.

        Returns:
            bool: True if withdrawal starts successfully, False otherwise.
        """
        if await self.is_pumping():
            logger.warning("Pump already moving!")

        if rate:  # Else previous rate will be used
            await self.hw_device.set_withdrawing_flow_rate(rate)

        if volume:  # FIXME check if target volume also works for withdrawing!
            await self.hw_device.set_target_volume(volume)

        return await self.hw_device.withdraw()



"""ML600 component relative to pumping."""
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from loguru import logger

from flowchem import ureg
from flowchem.components.pumps.syringe_pump import SyringePump
from flowchem.utils.exceptions import DeviceError

if TYPE_CHECKING:
    from .ml600 import ML600


class ML600Pump(SyringePump):
    """
    A component class representing an ML600 pump, capable of both infusion and withdrawal operations.

    Attributes:
    -----------
    pump_code : str
        Identifier for the pump ("" for single syringe pump, "B" or "C" for dual syringe pump).
    hw_device : ML600
        The hardware device instance associated with this component.

    Methods:
    --------
    is_withdrawing_capable() -> bool:
        Check if the pump supports withdrawal operations.
    is_pumping() -> bool:
        Check if the pump is currently moving.
    stop() -> bool:
        Stop the pump's operation.
    infuse(rate: str = "", volume: str = "") -> bool:
        Start an infusion with specified rate and volume.
    withdraw(rate: str = "1 ml/min", volume: str | None = None) -> bool:
        Start a withdrawal with specified rate and volume.
    """
    pump_code: str
    hw_device: ML600  # for typing's sake

    def __init__(self, name: str, hw_device: ML600, pump_code: str = "") -> None:
        """
        Initialize an ML600Pump object.

        Parameters:
        -----------
        name : str
            The name of the pump.
        hw_device : ML600
            The hardware device instance associated with this component.
        pump_code : str, optional
            Identifier for the pump (default is "", which denotes a single syringe pump).
            "" for single syringe pump. B or C  for dual syringe pump.
        """
        super().__init__(name, hw_device)
        self.pump_code = pump_code
        # self.add_api_route("/pump", self.get_monitor_position, methods=["GET"])

    @staticmethod
    def is_withdrawing_capable() -> bool:
        """
        Indicate that the ML600 pump can perform withdrawal operations.

        Returns:
        --------
        bool
            True, since ML600 supports withdrawal.
        """
        return True

    async def is_pumping(self) -> bool:
        """
        Check if the pump is currently moving.

        Returns:
        --------
        bool
            True if the pump is moving or has commands in buffer, False if it's idle.
        """
        # true might mean pump is moving, buffer still contain command or both
        return await self.hw_device.get_pump_status(self.pump_code)

    async def stop(self) -> bool:
        """
        Stop the pump's operation.

        Returns:
        --------
        bool
            True if the pump successfully stops, False otherwise.
        """
        await self.hw_device.stop(self.pump_code)
        # todo: sometime it take more then two seconds.
        await asyncio.sleep(1)
        if not await self.hw_device.get_pump_status(self.pump_code):
            return True
        else:
            logger.warning("the first check show false. try again.")
            await asyncio.sleep(1)
            return not await self.hw_device.get_pump_status(self.pump_code)

    async def infuse(self, rate: str = "", volume: str = "") -> bool:
        """
        Start an infusion with the given rate and volume.

        If no rate is specified, the default (1 ml/min) is used, can be set on per-pump basis via `default_infuse_rate`

        If no volume is specified, the max possible volume is infused.

        Parameters:
        -----------
        rate : str, optional
            The infusion rate (default is the device's configured default).
        volume : str, optional
            The volume to infuse (default is the maximum possible volume).

        Returns:
        --------
        bool
            True if the pump starts infusing successfully, False otherwise.

        Raises:
        -------
        DeviceError
            If the target volume to infuse exceeds the current syringe volume.
        """
        if await self.is_pumping():
            await self.stop()
        if not rate:
            rate = self.hw_device.config.get("default_infuse_rate")  # type: ignore
            logger.warning(f"the flow rate is not provided. set to the default {rate}")
        if not volume:
            target_vol = ureg.Quantity("0 ml")
            logger.warning("the volume to infuse is not provided. set to 0 ml")
        else:
            current_volume = await self.hw_device.get_current_volume(self.pump_code)
            target_vol = current_volume - ureg.Quantity(volume)
            if target_vol < 0:
                logger.error(
                    f"Cannot infuse target volume {volume}! "
                    f"Only {current_volume} in the syringe!",
                )
                raise DeviceError(f"Cannot infuse target volume {volume}! "
                                  f"Only {current_volume} in the syringe!")
                # return False

        await self.hw_device.set_to_volume(target_vol, ureg.Quantity(rate), self.pump_code)
        logger.info(f"infusing is run. it will take {ureg.Quantity(volume) / ureg.Quantity(rate)} to finish.")
        return await self.hw_device.get_pump_status(self.pump_code)

    async def withdraw(self, rate: str = "1 ml/min", volume: str | None = None) -> bool:
        """
        Start a withdrawal with the given rate and volume.

        The default can be set on per-pump basis via `default_withdraw_rate`.

        Parameters:
        -----------
        rate : str, optional
            The withdrawal rate (default is "1 ml/min").
        volume : str, optional
            The volume to withdraw (default is the maximum possible volume).

        Returns:
        --------
        bool
            True if the pump starts withdrawing successfully, False otherwise.

        Raises:
        -------
        DeviceError
            If the target volume to withdraw exceeds the syringe capacity.
        """
        if await self.is_pumping():
            await self.stop()
        if not rate:
            rate = self.hw_device.config["default_withdraw_rate"]
            logger.warning(f"the flow rate is not provided. set to the default {rate}")
        if volume is None:
            target_vol = self.hw_device.syringe_volume
            logger.warning(f"the volume to withdraw is not provided. set to {self.hw_device.syringe_volume}")
        else:
            current_volume = await self.hw_device.get_current_volume(self.pump_code)
            target_vol = current_volume + ureg.Quantity(volume)
            if target_vol > self.hw_device.syringe_volume:
                logger.error(
                    f"Cannot withdraw target volume {volume}! "
                    f"Max volume left is {self.hw_device.syringe_volume - current_volume}!",
                )
                raise DeviceError(f"Cannot withdraw target volume {volume}! "
                                  f"Max volume left is {self.hw_device.syringe_volume - current_volume}!")
                # return False

        await self.hw_device.set_to_volume(target_vol, ureg.Quantity(rate), self.pump_code)
        logger.info(f"withdrawing is run. it will take {ureg.Quantity(volume) / ureg.Quantity(rate)} to finish.")
        return await self.hw_device.get_pump_status(self.pump_code)



from __future__ import annotations

from pydantic import BaseModel


class ComponentInfo(BaseModel):
    """Metadata associated with flowchem components."""

    name: str = ""
    parent_device: str = ""
    type: str = ""
    corresponding_class: list[str] = []
    owl_subclass_of: list[str] = [
        "http://purl.obolibrary.org/obo/OBI_0000968",
    ]  # 'device'



"""FastAPI server for devices control."""
from collections.abc import Iterable
from importlib.metadata import metadata, version

from fastapi import APIRouter, FastAPI
from loguru import logger
from starlette.responses import RedirectResponse

from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import RepeatedTaskInfo
from flowchem.vendor.repeat_every import repeat_every


class FastAPIServer:
    def __init__(
        self, filename: str = "", host: str = "127.0.0.1", port: int = 8000
    ) -> None:
        # Create FastAPI app
        self.app = FastAPI(
            title=f"Flowchem - {filename}",
            description=metadata("flowchem")["Summary"],
            version=version("flowchem"),
            license_info={
                "name": "MIT License",
                "url": "https://opensource.org/licenses/MIT",
            },
        )
        self.base_url = rf"http://{host}:{port}"

        self._add_root_redirect()

        logger.debug("HTTP ASGI server app created")

    def _add_root_redirect(self) -> None:
        @self.app.route("/")
        def home_redirect_to_docs(request):
            """Redirect root to `/docs` to enable interaction w/ API."""
            return RedirectResponse(url="/docs")

    def add_background_tasks(self, repeated_tasks: Iterable[RepeatedTaskInfo]):
        """Schedule repeated tasks to run upon server startup."""
        for seconds_delay, task in repeated_tasks:
            @self.app.on_event("startup")
            @repeat_every(seconds=seconds_delay)
            async def my_task():
                logger.debug("Running repeated task...")
                await task()

    def add_device(self, device):
        """Add device to server."""
        # Add components URL to device_info
        components_w_url = {
            component.name: f"{self.base_url}/{device.name}/{component.name}"
            for component in device.components
        }
        device.device_info.components = components_w_url

        # Base device endpoint
        device_root = APIRouter(prefix=f"/{device.name}", tags=[device.name])
        device_root.add_api_route(
            "/",
            device.get_device_info,
            methods=["GET"],
            response_model=DeviceInfo,
        )
        self.app.include_router(device_root)

        # Add repeated tasks for device if any
        if tasks := device.repeated_task():
            self.add_background_tasks(tasks)

        # add device components
        logger.debug(f"Device '{device.name}' has {len(device.components)} components")
        for component in device.components:
            self.app.include_router(component.router, tags=component.router.tags)
            logger.debug(f"Router <{component.router.prefix}> added to app!")



"""Sensor device."""
from __future__ import annotations

from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.devices.flowchem_device import FlowchemDevice


class Sensor(FlowchemComponent):
    """
    A generic sensor device class.

    Attributes:
    -----------
    hw_device : FlowchemDevice
        The hardware device this sensor is interfacing with.

    Methods:
    --------
    power_on():
        Power on the sensor.
    power_off():
        Power off the sensor.
    """

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Constructs all the necessary attributes for the Sensor object.

        Parameters:
        -----------
        name : str
            The name of the sensor.
        hw_device : FlowchemDevice
            The hardware device this sensor is interfacing with.
        """
        super().__init__(name, hw_device)
        self.add_api_route("/power-on", self.power_on, methods=["PUT"])
        self.add_api_route("/power-off", self.power_off, methods=["PUT"])
        # Ontology: HPLC isocratic pump
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/NCIT_C50166",
        )

    async def power_on(self):
        """
        Power on the sensor.

        Returns:
        --------
        None
        """
        ...

    async def power_off(self):
        """
        Power off the sensor.

        Returns:
        --------
        None
        """
        ...



"""Control module for the Vapourtec R2."""
from __future__ import annotations

import asyncio
from asyncio import Lock
from collections import namedtuple
from collections.abc import Iterable

import aioserial
import pint
from loguru import logger

from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.components.technical.temperature import TempRange
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.vapourtec.r2_components_control import (
    R2GeneralPressureSensor,
    R2GeneralSensor,
    R2HPLCPump,
    R2InjectionValve,
    R2MainSwitch,
    R2PumpPressureSensor,
    R2TwoPortValve,
    R4Reactor,
    UV150PhotoReactor,
)
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import dario, jakob, wei_hsin

try:
    # noinspection PyUnresolvedReferences
    from flowchem_vapourtec import VapourtecR2Commands

    HAS_VAPOURTEC_COMMANDS = True
except ImportError:
    HAS_VAPOURTEC_COMMANDS = False


class R2(FlowchemDevice):
    """R2 reactor module class."""

    DEFAULT_CONFIG = {
        "timeout": 0.25,  # 0.22
        "baudrate": 19200,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }

    AllComponentStatus = namedtuple(
        "AllComponentStatus",
        [
            "run_state",
            "pumpA_speed",
            "pumpB_speed",
            "airlock1",
            "airlock2",
            "presslimit",
            "LEDs_bitmap",
            "chan1_temp",
            "chan2_temp",
            "chan3_temp",
            "chan4_temp",
            "U1",
            "U2",
            "U3",
            "L1",
            "L2",
        ],
    )

    def __init__(
        self,
        name: str = "",
        rt_temp: float = 27,  # todo: find a way to
        min_temp: float | list[float] = -40,
        max_temp: float | list[float] = 80,
        min_pressure: float = 1000,
        max_pressure: float = 50000,
        **config,
    ) -> None:
        super().__init__(name)

        # Set max pressure for R2 pump

        # Set min and max temp for R4 heater
        if not isinstance(min_temp, Iterable):
            min_temp = [min_temp] * 4
        if not isinstance(max_temp, Iterable):
            max_temp = [max_temp] * 4
        assert len(min_temp) == len(max_temp) == 4

        self.rt_t = rt_temp * ureg.degreeC
        self._min_t = min_temp * ureg.degreeC
        self._max_t = max_temp * ureg.degreeC

        self._heated = True  # to saving the dry ice
        self._intensity = 0

        if not HAS_VAPOURTEC_COMMANDS:
            raise InvalidConfigurationError(
                "You tried to use a Vapourtec device but the relevant commands are missing!\n"
                "Unfortunately, we cannot publish those as they were provided under NDA.\n"
                "Contact Vapourtec for further assistance."
            )

        self.cmd = VapourtecR2Commands()

        # Merge default settings, including serial, with provided ones.
        configuration = R2.DEFAULT_CONFIG | config
        try:
            self._serial = aioserial.AioSerial(**configuration)
        except aioserial.SerialException as ex:
            raise InvalidConfigurationError(
                f"Cannot connect to the R2 on the port <{config.get('port')}>"
            ) from ex

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Vapourtec",
            model="R2 reactor module",
        )
        self._serial_lock = Lock()

    async def initialize(self):
        """Ensure connection."""
        self.device_info.version = await self.version()
        logger.info(f"Connected with R2 version {self.device_info.version}")

        # Sets all pump to 0 ml/min
        await asyncio.sleep(0.1)
        await self.set_flowrate("A", "0 ul/min")
        await self.set_flowrate("B", "0 ul/min")
        # Sets all temp to room temp.
        # rt = ureg("25 °C")
        self.rt_t = await self.get_current_temperature(2) * ureg.degreeC
        await self.set_temperature(0, self.rt_t, self._heated)
        await self.set_temperature(1, self.rt_t, self._heated)
        await self.set_temperature(2, self.rt_t, self._heated)
        await self.set_temperature(3, self.rt_t, self._heated)

        # set UV to 0%
        await self.set_UV150(power=self._intensity)
        # set max pressure to  10 bar
        await self.set_pressure_limit("20 bar")
        # Set valve to default position
        await self.trigger_key_press("0")
        await self.trigger_key_press("2")
        await self.trigger_key_press("4")
        await self.trigger_key_press("6")
        await self.trigger_key_press("8")
        await self.power_on()

        list_of_components = [
            R2MainSwitch("Power", self),
            R2GeneralPressureSensor("PressureSensor", self),
            R2GeneralSensor("GSensor2", self),
            UV150PhotoReactor("PhotoReactor", self),
            R2HPLCPump("Pump_A", self, "A"),
            R2HPLCPump("Pump_B", self, "B"),
            R2TwoPortValve("ReagentValve_A", self, 0),
            R2TwoPortValve("ReagentValve_B", self, 1),
            R2TwoPortValve("CollectionValve", self, 4),
            R2InjectionValve("InjectionValve_A", self, 2),
            R2InjectionValve("InjectionValve_B", self, 3),
            R2PumpPressureSensor("PumpSensor_A", self, 0),
            R2PumpPressureSensor("PumpSensor_B", self, 1),
        ]
        self.components.extend(list_of_components)

        # TODO if photoreactor -> REACTOR CHANNEL 1,3 AND 4 + UV150PhotoReactor
        #  if no photoreactor -> REACTOR CHANNEL 1-4 no UV150PhotoReactor

        # Create components for reactor bays
        reactor_temp_limits = {
            ch_num: TempRange(min=ureg.Quantity(t[0]), max=ureg.Quantity(t[1]))
            for ch_num, t in enumerate(zip(self._min_t, self._max_t, strict=True))
        }

        reactors = [
            R4Reactor(f"reactor-{n + 1}", self, n, reactor_temp_limits[n])
            for n in range(4)
        ]
        self.components.extend(reactors)

    async def _write(self, command: str):
        """Write a command to the pump."""
        cmd = command + "\r\n"
        await self._serial.write_async(cmd.encode("ascii"))
        logger.debug(f"Sent command: {command!r}")

    async def _read_reply(self) -> str:
        """Read the pump reply from serial communication."""
        reply_string = await self._serial.readline_async()
        logger.debug(f"Reply received: {reply_string.decode('ascii').rstrip()}")
        return reply_string.decode("ascii")

    async def write_and_read_reply(self, command: str) -> str:
        """Send a command to the pump, read the replies and return it, optionally parsed."""
        self._serial.reset_input_buffer()  # Clear input buffer, discarding all that is in the buffer.
        async with self._serial_lock:
            await self._write(command)
            logger.debug(f"Command {command} sent to R2!")

            failure = 0
            while True:
                response = await self._read_reply()
                if not response:
                    failure += 1
                    logger.warning(f"{failure} time of failure!")
                    logger.error(f"Command {command} is not working")
                    await asyncio.sleep(0.2)
                    self._serial.reset_input_buffer()
                    await self._write(command)
                    # Allows 4 failures...
                    if failure > 3:
                        raise InvalidConfigurationError(
                            "No response received from R2 module!"
                        )
                else:
                    break

        logger.debug(f"Reply received: {response}")
        return response.rstrip()

    # Specific Commands
    # Get instrument information
    async def version(self):
        """Get firmware version."""
        return await self.write_and_read_reply(self.cmd.VERSION)

    async def system_type(self):
        """Get system type: system type, pressure mode."""
        return await self.write_and_read_reply(self.cmd.GET_SYSTEM_TYPE)

    async def get_status(self) -> AllComponentStatus:
        """Get all status from R2."""
        raw_status = await self.write_and_read_reply(self.cmd.GET_STATUS)
        if raw_status == "OK":
            logger.warning("ValueError:the reply of get status command is OK..")
            return await self.get_status()
        return R2.AllComponentStatus._make(raw_status.split(" "))

    # Get specific state of individual component
    async def get_state(self) -> str:
        """Get run state."""
        State_dic = {
            "0": "Off",
            "1": "Running",
            "2": "System overpressure",
            "3": "Pump A overpressure",
            "4": "Pump B overpressure",
            "5": "System pressure loss",
            "6": "Pump A pressure loss",
            "7": "Pump B pressure loss",
        }
        state = await self.get_status()
        return State_dic[state.run_state]

    async def get_setting_Pressure_Limit(self) -> str:
        """Get system pressure limit."""
        state = await self.get_status()
        return state.presslimit

    async def get_target_temperature(self, channel) -> float:
        """Get temperature (in Celsius) from R4."""
        state = await self.get_status()
        temp_list = [
            state.chan1_temp,
            state.chan2_temp,
            state.chan3_temp,
            state.chan4_temp,
        ]
        return float(temp_list[channel])

    async def get_valve_position(self, valve_code: int) -> str:
        """Get specific valves position."""
        state = await self.get_status()
        # Current state of all valves as bitmap
        bitmap = int(state.LEDs_bitmap)

        return list(reversed(f"{bitmap:05b}"))[valve_code]

    # Set parameters
    async def set_flowrate(self, pump: str, flowrate: str):
        """Set flow rate to pump ('A'|'B')."""
        if flowrate.isnumeric():
            flowrate = flowrate + "ul/min"
            logger.warning(
                "No units provided to set_temperature, assuming microliter/minutes.",
            )
        parsed_f = ureg.Quantity(flowrate)

        if pump == "A":
            pump_num = 0
        elif pump == "B":
            pump_num = 1
        else:
            logger.warning(f"Invalid pump name: {pump}")
            return

        cmd = self.cmd.SET_FLOWRATE.format(
            pump=pump_num,
            rate_in_ul_min=round(parsed_f.m_as("ul/min")),
        )
        await self.write_and_read_reply(cmd)

    async def set_temperature(
        self,
        channel: int,
        temp: pint.Quantity,
        heating: bool | None = None,
        ramp_rate: str = "80",
    ):
        """Set temperature to R4 channel. If a UV150 is present then channel 3 range is limited to -40 to 80 °C.
        The heating setting range from 20 to 80 °C; cooling range from -40 to 80 °C.
        """
        set_t = round(temp.m_as("°C"))

        if heating is None:
            logger.debug("heat decide by temp.")
            # todo: better heat decision
            # For 525 nm green light (13W) the temp difference to rt is 8 degree; 440 nm blue light (24W) ~15 degree
            threhold_t = self.rt_t.m_as("degree_Celsius") + 8
            self._heated = True if set_t > threhold_t else False
        elif heating is True:
            self._heated = True
        else:
            self._heated = False

        # change the setting of heating
        cmd = self.cmd.SET_UV150.format(
            power_percent=self._intensity, heater_on=int(self._heated)
        )
        await self.write_and_read_reply(cmd)

        # set the temperature
        cmd = self.cmd.SET_TEMPERATURE.format(
            channel=channel,
            temperature_in_C=set_t,
            ramp_rate=ramp_rate,
        )
        await self.write_and_read_reply(cmd)

    async def set_pressure_limit(self, pressure: str):
        """Set maximum system pressure: range 1,000 to 50,000 mbar."""
        if pressure.isnumeric():
            pressure = pressure + "mbar"
            logger.warning("No units provided to set_temperature, assuming mbar.")
        set_p = ureg.Quantity(pressure)

        cmd = self.cmd.SET_MAX_PRESSURE.format(
            max_p_in_mbar=round(set_p.m_as("mbar") / 500) * 500,
        )
        await self.write_and_read_reply(cmd)

    async def set_UV150(self, power: int):
        """set intensity of the UV light: 0 or 50 to 100"""
        self._intensity = power
        cmd = self.cmd.SET_UV150.format(
            power_percent=self._intensity, heater_on=int(self._heated)
        )
        await self.write_and_read_reply(cmd)

    async def trigger_key_press(self, keycode: str):
        """Set valve position."""
        cmd = self.cmd.KEY_PRESS.format(keycode=keycode)
        await self.write_and_read_reply(cmd)

    async def power_on(self):
        """Turn on both devices, R2 and R4."""
        await self.write_and_read_reply(self.cmd.POWER_ON)

    async def power_off(self):
        """Turn off both devices, R2 and R4."""
        await self.write_and_read_reply(self.cmd.POWER_OFF)

    async def get_current_temperature(self, channel) -> float:
        """Get temperature (in Celsius) from channel3."""
        temp_history = await self.write_and_read_reply(self.cmd.HISTORY_TEMPERATURE)
        if temp_history == "OK":
            logger.warning("ValueError:the reply of get temperature command is OK..")
            return await self.get_current_temperature(channel)

        # 0: time, 1..8: cooling, heating, or ...  / temp (alternating per channel)
        temp_time, *temps = temp_history.split(",")
        return float(temps[channel * 2 + 1]) / 10

    async def get_pressure_history(
        self,
    ) -> tuple[int, int, int]:
        """Get pressure history and returns it as (in mbar)."""
        # Get a `&` separated list of pressures for all sensors every second
        pressure_history = await self.write_and_read_reply(self.cmd.HISTORY_PRESSURE)
        if pressure_history == "OK":
            logger.warning("ValueError:the reply for get pressure command is OK....")
            # This may give an infinite loop
            return await self.get_pressure_history()
        # Each pressure data point consists of four values: time and three pressures
        _, *pressures = pressure_history.split("&")[0].split(
            ",",
        )  # e.g. 45853,94,193,142
        # Converts to mbar
        p_in_mbar = [int(x) * 10 for x in pressures]
        return p_in_mbar[1], p_in_mbar[2], p_in_mbar[0]  # pumpA, pumpB, system

    async def get_current_pressure(self, pump_code: int = 2) -> pint.Quantity:
        """Get current pressure (in mbar)."""
        press_state_list = await self.get_pressure_history()
        # 0: pump A, 1: pump_B, 2: system pressure
        return press_state_list[pump_code] * ureg.mbar

    async def get_current_flow(self, pump_code: str) -> float:
        """Get current flow rate (in ul/min)."""
        state = await self.write_and_read_reply(self.cmd.HISTORY_FLOW)
        if state == "OK":
            logger.warning("ValueError:the reply of get flow command is OK....")
            return await self.get_current_flow(pump_code)

        pump_flow = {}
        # _: time, A: pump A, B: pump B
        _, pump_flow["A"], pump_flow["B"] = state.split("&")[0].split(",")
        # 0: time, 1: pump A, 2: pump B
        return float(pump_flow[pump_code])

    async def pooling(self) -> dict:
        """Extract all reaction parameters."""
        AllState = {}
        while True:
            state = await self.get_status()
            AllState["RunState_code"] = state.run_state
            AllState["allValve"] = f"{int(state.LEDs_bitmap):05b}"
            (
                AllState["pumpA_P"],
                AllState["pumpB_P"],
                AllState["sysP (mbar)"],
            ) = await self.get_pressure_history()

            # AllState["Temp"] = await self.get_current_temperature()
            return AllState


if __name__ == "__main__":
    R2_device = R2(port="COM6")

    async def main(Vapourtec_R2):
        """Test function."""
        await Vapourtec_R2.initialize()
        # Get valve and pump
        (
            r2swich,
            sysPs,
            Gs,
            pr2,
            pA,
            pB,
            vA,
            vB,
            vC,
            ivA,
            ivB,
            sA,
            sB,
            r1,
            r2,
            r3,
            r4,
        ) = Vapourtec_R2.components()

        print(f"{await Vapourtec_R2.pooling()}")
        # print(f"Injection valve A {await ivA.get_monitor_position()}")
        print(await Vapourtec_R2.get_valve_position(1))
        # print(f"The system state is {await Vapourtec_R2.get_Run_State()}")
        # await r3.set_temperature("30°C")
        # print(f"Temperature is {await Vapourtec_R2.get_target_temperature(2)}")
        # await pA.set_flowrate("100 ul/min")
        # while True:
            # await pA.infuse("10 ul/min")
            # print(f"{await Vapourtec_R2.pooling()}")

            # print(f"current pressure of pump A is {await sA.read_pressure()}")

            # print(f"{await Vapourtec_R2.get_valve_position(2)}")
            # print(f"Injection valve A {await ivA.get_position()}")
            # await ivA.set_position("inject")
            # await asyncio.sleep(0.5)
            # print(f"Injection valve A {await ivA.get_position()}")
        await r2swich.power_off()

    asyncio.run(main(R2_device))



from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.technical.power import PowerSwitch
from flowchem.devices.flowchem_device import FlowchemDevice

if TYPE_CHECKING:
    from .bubble_sensor import PhidgetBubbleSensor, PhidgetPowerSource5V

from flowchem.components.sensors.sensor import Sensor


class PhidgetBubbleSensorComponent(Sensor):
    hw_device: PhidgetBubbleSensor  # just for typing

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(name, hw_device)
        # self.add_api_route("/set-data-Interval", self.set_dataInterval, methods=["PUT"])
        self.add_api_route("/read-voltage", self.read_voltage, methods=["GET"])
        self.add_api_route("/acquire-signal", self.acquire_signal, methods=["GET"])

    async def power_on(self) -> bool:
        self.hw_device.power_on()
        return True

    async def power_off(self) -> bool:
        self.hw_device.power_off()
        return True

    async def read_voltage(self) -> float:
        """Read from sensor in Volt."""
        return self.hw_device.read_voltage()

    async def acquire_signal(self) -> float:
        """Transform the voltage from sensor to be expressed in percentage(%)."""
        return self.hw_device.read_intensity()

    async def set_dataInterval(self, datainterval: int) -> bool:
        """Set data interval at the range 20-60000 ms (default unit: ms)."""
        self.hw_device.set_dataInterval(datainterval)
        return True


class PhidgetBubbleSensorPowerComponent(PowerSwitch):
    hw_device: PhidgetPowerSource5V  # just for typing

    async def power_on(self) -> bool:
        self.hw_device.power_on()
        return True

    async def power_off(self) -> bool:
        self.hw_device.power_off()
        return True



"""Autodiscover Knauer devices on network."""
import asyncio
import queue
import socket
import sys
from textwrap import dedent

import ifaddr
from anyio.from_thread import start_blocking_portal
from loguru import logger

from flowchem.vendor.getmac import get_mac_address

__all__ = ["autodiscover_knauer", "knauer_finder"]

Address = tuple[str, int]


class BroadcastProtocol(asyncio.DatagramProtocol):
    """See `https://gist.github.com/yluthu/4f785d4546057b49b56c`."""

    def __init__(self, target: Address, response_queue: queue.Queue) -> None:
        self.target = target
        self.loop = asyncio.get_event_loop()
        self._queue = response_queue

    def connection_made(self, transport: asyncio.transports.DatagramTransport):  # type: ignore
        """Send the magic broadcast package for autodiscovery."""
        sock = transport.get_extra_info("socket")  # type: socket.socket
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)  # sets to broadcast
        transport.sendto(b"\x00\x01\x00\xf6", self.target)

    def datagram_received(self, data: bytes | str, addr: Address):
        """Add received data to queue."""
        logger.trace(f"Received data from {addr}")
        self._queue.put(addr[0])


async def get_device_type(ip_address: str) -> str:
    """Detect device type based on the reply to a test command or IP heuristic."""
    fut = asyncio.open_connection(host=ip_address, port=10001)
    try:
        reader, writer = await asyncio.wait_for(fut, timeout=3)
    except ConnectionError:
        return "ConnectionError"
    except asyncio.TimeoutError:
        if ip_address == "192.168.1.2":
            return "TimeoutError - Nice FlowIR that you have :D"
        return "TimeoutError"

    # Test Pump
    writer.write(b"HEADTYPE:?\n\r")
    try:
        reply = await asyncio.wait_for(reader.readuntil(separator=b"\r"), timeout=1)
    except asyncio.TimeoutError:
        pass
    else:
        if reply.startswith(b"HEADTYPE"):
            logger.debug(f"Device {ip_address} is a pump")
            return "AzuraCompact"
    logger.debug("Not a pump")

    # Test Valve
    writer.write(b"T:?\n\r")
    try:
        reply = await asyncio.wait_for(reader.readuntil(separator=b"\r"), timeout=1)
    except asyncio.TimeoutError:
        pass
    else:
        if reply.startswith(b"VALVE"):
            logger.debug(f"Device {ip_address} is a valve")
            return "KnauerValve"
    logger.debug("Not a valve")

    return "Unknown"


async def send_broadcast_and_receive_replies(source_ip: str):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    device_q: queue.Queue = queue.Queue()
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: BroadcastProtocol(("255.255.255.255", 30718), response_queue=device_q),
        local_addr=(source_ip, 28688),
        allow_broadcast=True,
    )
    try:
        await asyncio.sleep(2)
    finally:
        transport.close()

    device_list = []
    # Get all device from queue (nobody should need has more than 40 devices, right?)
    for _ in range(40):
        try:
            device_list.append(device_q.get_nowait())
        except queue.Empty:
            break

    return device_list


def broadcast_ip_heuristic(ip: str) -> int:
    """Heuristic to determine the preferred IP to be used in broadcasting to find local devices."""
    if ip.startswith("192.168"):
        return 0
    if ip.startswith("10."):
        return 1
    if ip.startswith("100."):
        return 2
    if ip.startswith("141."):
        return 3
    if ip.startswith("127.0"):
        return 100
    return 10


def determine_broadcasting_ip(source="") -> str:
    """Determine the broadcasting IP to be used based on NIC name or IP range (e.g. 'eth0' or '192.168.*.*')."""
    assert isinstance(source, str), "Source must be a string"
    # Get network interfaces data
    nics = {}
    for adapter in ifaddr.get_adapters():
        for ip in adapter.ips:
            # Skip IPv6
            if isinstance(ip.ip, tuple):
                continue
            # We have an IPv4 as string
            nics[ip.nice_name] = ip.ip

    # If a valid IPv4 was provided use that
    if source in nics.values():
        logger.debug(f"Broadcasting IP: {source} [passed from settings].")
        logger.info(
            f"Consider specifying the source IP as one of these: {list(nics.keys())}"
        )
        return source

    # If a NIC is named as source return the corresponding interface IP
    try:
        logger.debug(f"Broadcasting IP: {nics[source]} [based on NIC: {source}]")
        return nics[source]
    except KeyError:
        pass
        # logger.debug(f"No network interface '{source}' found. Available are: {list(nics.keys())}")

    # If the source IP is not fully defined try to find a match among all the local IPs available
    if "*" in source:
        fixed_part, *_ = source.split("*")

        for nic_ip in nics.values():
            if nic_ip.startswith(fixed_part):
                logger.debug(f"Broadcasting IP: {nic_ip} [based on IP range: {source}]")
                return nic_ip

    # Find IP
    if source:
        logger.warning(f"IP source for broadcasting invalid! ['{source}' provided]")

    # Get one of the local IPs and hope for the best.
    return sorted(nics.values(), key=broadcast_ip_heuristic)[0]


def autodiscover_knauer(network: str = "") -> dict[str, str]:
    """
    Automatically find Knauer ethernet device on the network and returns the IP associated to each MAC address.
    Note that the MAC is the key here as it is the parameter used in configuration files.
    Knauer devices only support DHCP so static IPs are not an option.

    Args:
    ----
        network: source for autodiscover (either IP, NIC or IP range e.g. '192.168.*.*')
    Returns:
    -------
        List of tuples (IP, MAC, device_type), one per device replying to autodiscover.
    """
    source_ip = determine_broadcasting_ip(network)
    logger.info(f"Starting detection from IP {source_ip}")

    with start_blocking_portal() as portal:
        device_list = portal.call(send_broadcast_and_receive_replies, source_ip)

    device_info: dict[str, str] = {}
    device_ip: str
    # We got replies from IPs, let's find their MACs
    for device_ip in device_list:
        logger.debug(f"Got a reply from {device_ip}")
        # MAC address
        mac = get_mac_address(ip=device_ip)
        if mac:
            device_info[mac] = device_ip
    return device_info


def knauer_finder(source_ip: str = ""):
    """Execute autodiscovery. This is the entry point of the `knauer-finder` CLI command."""
    # This is a bug of asyncio on Windows :|
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    # Autodiscover devices (returns dict with MAC as index, IP as value)
    devices = autodiscover_knauer(source_ip)
    dev_config = set()

    for mac_address, ip in devices.items():
        logger.info(f"Determining device type for device at {ip} [{mac_address}]")
        # Device Type
        device_type = asyncio.run(get_device_type(ip))
        logger.info(f"Device type detected for IP {ip}: {device_type}")

        match device_type:
            case "AzuraCompact":
                dev_config.add(
                    dedent(
                        f"""
                        [device.pump-{mac_address[-8:-6] + mac_address[-5:-3] + mac_address[-2:]}]
                        type = "AzuraCompact"
                        ip_address = "{ip}"  # MAC address during discovery: {mac_address}
                        # max_pressure = "XX bar"
                        # min_pressure = "XX bar"\n\n""",
                    ),
                )
            case "KnauerValve":
                dev_config.add(
                    dedent(
                        f"""
                        [device.valve-{mac_address[-8:-6] + mac_address[-5:-3] + mac_address[-2:]}]
                        type = "KnauerValve"
                        ip_address = "{ip}"  # MAC address during discovery: {mac_address}\n\n""",
                    ),
                )
            case "FlowIR":
                dev_config.add(
                    dedent(
                        """
                        [device.flowir]
                        type = "IcIR"
                        url = "opc.tcp://localhost:62552/iCOpcUaServer"  # Default, replace with IP of PC with IcIR
                        template = "some-template.iCIRTemplate"  # Replace with valid template name, see docs.\n\n""",
                    ),
                )

    return dev_config


if __name__ == "__main__":
    knauer_finder()



"""Autodiscover iCIR opcua server locally."""
import asyncio
import sys
from textwrap import dedent

from loguru import logger

from flowchem.devices import IcIR

__all__ = ["icir_finder"]

Address = tuple[str, int]


async def is_iCIR_running_locally() -> bool:
    """Is iCIR available on the local machine (default URL)?."""
    ir = IcIR()
    try:
        await ir.opcua.connect()
    except asyncio.TimeoutError:
        return False

    return await ir.is_iCIR_connected()


async def generate_icir_config() -> str:
    """Generate config string if iCIR is available."""
    if await is_iCIR_running_locally():
        logger.debug("Local iCIR found!")
        return dedent(
            """
               [device.icir-local]
               type = "IcIR"
               template = ""  # Add template name with acquisition settings!
               \n\n""",
        )
    return ""


def icir_finder():
    """Attempt connection on local iCIR instance."""
    # This is a bug of asyncio on Windows :|
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    try:
        return [asyncio.run(generate_icir_config())]
    except OSError:
        return []


if __name__ == "__main__":
    print(icir_finder())



"""Async implementation of FlowIR."""
import asyncio
import datetime
from pathlib import Path

from asyncua import Client, ua
from asyncua.ua.uaerrors import BadOutOfService, Bad
from loguru import logger
from pydantic import BaseModel

from flowchem.components.analytics.ir import IRSpectrum
from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.mettlertoledo.icir_control import IcIRControl
from flowchem.utils.exceptions import DeviceError
from flowchem.utils.people import dario, jakob, wei_hsin


class ProbeInfo(BaseModel):
    """Dictionary returned from iCIR with probe info."""

    spectrometer: str
    spectrometer_SN: int
    probe_SN: int
    detector: str
    apodization: str
    ip_address: str
    probe_type: str
    sampling_interval: str
    resolution: int
    scan_option: str
    gain: int


class IcIR(FlowchemDevice):
    """Object to interact with the iCIR software controlling the FlowIR and ReactIR."""

    iC_OPCUA_DEFAULT_SERVER_ADDRESS = "opc.tcp://localhost:62552/iCOpcUaServer"
    _supported_versions = {"7.1.91.0"}
    SOFTWARE_VERSION = "ns=2;s=Local.iCIR.SoftwareVersion"
    CONNECTION_STATUS = "ns=2;s=Local.iCIR.ConnectionStatus"
    PROBE_DESCRIPTION = "ns=2;s=Local.iCIR.Probe1.ProbeDescription"
    PROBE_STATUS = "ns=2;s=Local.iCIR.Probe1.ProbeStatus"
    LAST_SAMPLE_TIME = "ns=2;s=Local.iCIR.Probe1.LastSampleTime"
    SAMPLE_COUNT = "ns=2;s=Local.iCIR.Probe1.SampleCount"
    SPECTRA_TREATED = "ns=2;s=Local.iCIR.Probe1.SpectraTreated"
    SPECTRA_RAW = "ns=2;s=Local.iCIR.Probe1.SpectraRaw"
    SPECTRA_BACKGROUND = "ns=2;s=Local.iCIR.Probe1.SpectraBackground"
    START_EXPERIMENT = "ns=2;s=Local.iCIR.Probe1.Methods.Start Experiment"
    STOP_EXPERIMENT = "ns=2;s=Local.iCIR.Probe1.Methods.Stop"
    METHODS = "ns=2;s=Local.iCIR.Probe1.Methods"

    counter = 0

    def __init__(self, template="", url="", name="") -> None:
        """Initiate connection with OPC UA server."""
        super().__init__(name)
        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Mettler-Toledo",
            model="iCIR",
            version="",
        )

        # Default (local) url if none provided
        if not url:
            url = self.iC_OPCUA_DEFAULT_SERVER_ADDRESS
        self.opcua = Client(
            url,
            timeout=5,
        )  # Call to START_EXPERIMENT can take few seconds!

        self._template = template

    async def initialize(self):
        """Initialize, check connection and start acquisition."""
        try:
            await self.opcua.connect()
        except asyncio.TimeoutError as timeout_error:
            raise DeviceError(
                f"Could not connect to iCIR on {self.opcua.server_url}!"
            ) from timeout_error

        # Ensure iCIR version is supported
        self.device_info.version = await self.opcua.get_node(
            self.SOFTWARE_VERSION,
        ).get_value()  # e.g. "7.1.91.0"

        self.ensure_version_is_supported()
        logger.debug("iCIR initialized!")

        if not await self.is_iCIR_connected():
            raise DeviceError("Device not connected! Check iCIR...")

        # Start acquisition! Ensures the device is ready when a spectrum is needed
        await self.start_experiment(name="Flowchem", template=self._template)
        probe = await self.probe_info()
        self.device_info.additional_info = probe.model_dump()

        # Set IRSpectrometer component
        self.components.append(IcIRControl("ir-control", self))

    def is_local(self):
        """Return true if the server is on the same machine running the python code."""
        return any(
            x in self.opcua.server_url.netloc for x in ("localhost", "127.0.0.1")
        )

    def ensure_version_is_supported(self):
        """Check if iCIR is installed and open and if the version is supported."""
        try:
            if self.device_info.version not in self._supported_versions:
                logger.warning(
                    f"The current version of iCIR [self.version] has not been tested!"
                    f"Pleas use one of the supported versions: {self._supported_versions}",
                )
        except ua.UaStatusCodeError as error:  # iCIR app closed
            raise DeviceError(
                "iCIR app not installed or closed or no instrument available!"
            ) from error

    # noinspection PyPep8Naming
    async def is_iCIR_connected(self) -> bool:
        """Check connection with instrument."""
        return await self.opcua.get_node(self.CONNECTION_STATUS).get_value()

    async def probe_info(self) -> ProbeInfo:
        """Return FlowIR probe information."""
        probe_info = await self.opcua.get_node(self.PROBE_DESCRIPTION).get_value()
        return self.parse_probe_info(probe_info)

    async def probe_status(self):
        """Return current probe status. Possible values are 'Running', 'Not running' (+ more?)."""
        return await self.opcua.get_node(self.PROBE_STATUS).get_value()

    async def last_sample_time(self) -> datetime.datetime:
        """Return date/time of the latest scan."""
        return await self.opcua.get_node(self.LAST_SAMPLE_TIME).get_value()

    async def sample_count(self) -> int | None:
        """Sample count (integer autoincrement) watch for changes to ensure latest spectrum is recent."""
        return await self.opcua.get_node(self.SAMPLE_COUNT).get_value()

    @staticmethod
    def _normalize_template_name(template_name) -> str:
        """Add `.iCIRTemplate` extension to string if not already present."""
        return (
            template_name
            if template_name.endswith(".iCIRTemplate")
            else template_name + ".iCIRTemplate"
        )

    @staticmethod
    def is_template_name_valid(template_name: str) -> bool:
        r"""Check template name validity. For the template folder location read below.

        From Mettler Toledo docs:
        You can use the Start method to create and run a new experiment in one of the iC analytical applications
        (i.e. iC IR, iC FBRM, iC Vision, iC Raman). Note that you must provide the name of an existing experiment
        template file that can be used as a basis for the new experiment.
        The template file must be located in a specific folder on the iC OPC UA Server computer.
        This is usually C:\\ProgramData\\METTLER TOLEDO\\iC OPC UA Server\\1.2\\Templates.
        """
        template_dir = Path(
            r"C:\ProgramData\METTLER TOLEDO\iC OPC UA Server\1.2\Templates",
        )
        if not template_dir.exists() or not template_dir.is_dir():
            logger.warning("iCIR template folder not found on the local PC!")
            return False

        # Ensures the name has been provided with no extension (common mistake)
        template_name = IcIR._normalize_template_name(template_name)

        return any(
            existing_tmpl.name == template_name
            for existing_tmpl in template_dir.glob("*.iCIRTemplate")
        )

    @staticmethod
    def parse_probe_info(probe_info_reply: str) -> ProbeInfo:
        """Convert the device reply into a ProbeInfo dictionary.

        Example probe_info_reply reply is:
        'FlowIR; SN: 2989; Detector: DTGS; Apodization: HappGenzel; IP Address: 192.168.1.2;
        Probe: DiComp (Diamond); SN: 14570173; Interface: FlowIR™ Sensor; Sampling: 4000 to 650 cm-1;
        Resolution: 8; Scan option: AutoSelect; Gain: 232;'
        """
        fields = probe_info_reply.split(";")
        probe_info = {
            "spectrometer": fields[0],
            "spectrometer_SN": fields[1].split(": ")[1],
            "probe_SN": fields[6].split(": ")[1],
        }

        # Use aliases, i.e. translate API names (left) to dict key (right)
        translate_attributes = {
            "Detector": "detector",
            "Apodization": "apodization",
            "IP Address": "ip_address",
            "Probe": "probe_type",
            "Sampling": "sampling_interval",
            "Resolution": "resolution",
            "Scan option": "scan_option",
            "Gain": "gain",
        }
        for element in fields:
            if ":" in element:
                piece = element.split(":")
                if piece[0].strip() in translate_attributes:
                    probe_info[translate_attributes[piece[0].strip()]] = piece[
                        1
                    ].strip()

        return ProbeInfo.model_validate(probe_info)

    @staticmethod
    async def _wavenumber_from_spectrum_node(node) -> list[float]:
        """Get the X-axis value of a spectrum. This is necessary as they change e.g. with resolution."""
        node_property = await node.get_properties()
        x_axis = await node_property[0].get_value()
        return x_axis.AxisSteps

    @staticmethod
    async def spectrum_from_node(node) -> IRSpectrum:
        """Given a Spectrum node returns it as IRSpectrum."""
        try:
            intensity = await node.get_value()
            wavenumber = await IcIR._wavenumber_from_spectrum_node(node)
            return IRSpectrum(wavenumber=wavenumber, intensity=intensity)

        except BadOutOfService:
            return IRSpectrum(wavenumber=[], intensity=[])

    async def last_spectrum_treated(self) -> IRSpectrum:
        """Return an IRSpectrum element for the last acquisition."""
        return await IcIR.spectrum_from_node(self.opcua.get_node(self.SPECTRA_TREATED))

    async def last_spectrum_raw(self) -> IRSpectrum:
        """RAW result latest scan."""
        return await IcIR.spectrum_from_node(self.opcua.get_node(self.SPECTRA_RAW))

    async def last_spectrum_background(self) -> IRSpectrum:
        """RAW result latest scan."""
        return await IcIR.spectrum_from_node(
            self.opcua.get_node(self.SPECTRA_BACKGROUND),
        )

    async def start_experiment(
        self,
        template: str,
        name: str = "Unnamed flowchem exp.",
    ):
        r"""Start an experiment on iCIR.

        Args:
        ----
            template: name of the experiment template, should be in the Templtates folder on the PC running iCIR.
                      That usually is C:\\ProgramData\\METTLER TOLEDO\\iC OPC UA Server\1.2\\Templates
            name: experiment name.
        """
        template = self._normalize_template_name(template)
        if self.is_local() and self.is_template_name_valid(template) is False:
            raise DeviceError(
                f"Cannot start template {template}: name not valid! Check if is in: "
                r'"C:\ProgramData\METTLER TOLEDO\iC OPC UA Server\1.2\Templates"'
            )
        if await self.probe_status() == "Running":
            logger.warning(
                "I was asked to start an experiment while a current experiment is already running!"
                "I will have to stop that first! Sorry for that :)",
            )
            # Stop running experiment and wait for the spectrometer to be ready
            await self.stop_experiment()
            await self.wait_until_idle()

        start_xp_nodeid = self.opcua.get_node(self.START_EXPERIMENT).nodeid
        method_parent = self.opcua.get_node(self.METHODS)
        try:
            # Collect_bg does not seem to work in automation, set to false and do not expose in start_experiment()!
            collect_bg = False
            await method_parent.call_method(start_xp_nodeid, name, template, collect_bg)
        except Bad as error:
            raise DeviceError(
                "The experiment could not be started!\n"
                "Check iCIR status and close any open experiment."
            ) from error
        logger.info(f"FlowIR experiment {name} started with template {template}!")
        return True

    async def stop_experiment(self):
        """Stop the experiment currently running.

        Note: the call does not make the instrument idle: you need to wait for the current scan to end!
        """
        method_parent = self.opcua.get_node(self.METHODS)
        stop_nodeid = self.opcua.get_node(self.STOP_EXPERIMENT).nodeid
        await method_parent.call_method(stop_nodeid)

    async def wait_until_idle(self):
        """Wait until no experiment is running."""
        while await self.probe_status() == "Running":
            await asyncio.sleep(0.2)


if __name__ == "__main__":
    ...

    async def main():
        template_name = "15_sec_integration.iCIRTemplate"
        url = IcIR.iC_OPCUA_DEFAULT_SERVER_ADDRESS.replace(
            "localhost", "BSMC-YMEF002121"
        )
        ir_spectrometer = IcIR(template_name, url)

        if await ir_spectrometer.is_iCIR_connected():
            print("FlowIR connected!")
        else:
            raise ConnectionError("FlowIR not connected :(")

        await ir_spectrometer.start_experiment(
            name="reaction_monitoring", template=template_name
        )

        spectrum = await ir_spectrometer.last_spectrum_treated()
        while len(spectrum.intensity) == 0:
            spectrum = await ir_spectrometer.last_spectrum_treated()

        for x in range(3):
            spectra_count = await ir_spectrometer.sample_count()

            while await ir_spectrometer.sample_count() == spectra_count:
                await asyncio.sleep(1)

            print("New spectrum!")
            spectrum = await ir_spectrometer.last_spectrum_treated()
            print(spectrum)

        await ir_spectrometer.stop_experiment()

    asyncio.run(main())



import requests
from loguru import logger
from pydantic import AnyHttpUrl

from flowchem.client.component_client import FlowchemComponentClient
from flowchem.components.device_info import DeviceInfo


class FlowchemDeviceClient:
    def __init__(self, url: AnyHttpUrl) -> None:
        self.url = str(url)

        # Log every request and always raise for status
        self._session = requests.Session()
        self._session.hooks["response"] = [
            FlowchemDeviceClient.raise_for_status,
            FlowchemDeviceClient.log_responses,
        ]

        # Connect, get device info and populate components
        try:
            self.device_info = DeviceInfo.model_validate_json(
                self._session.get(self.url).text,
            )
        except ConnectionError as ce:
            msg = (
                f"Cannot connect to device at {url}!"
                f"This is likely caused by the server listening only on local the interface,"
                f"start flowchem with the --host 0.0.0.0 option to check if that's the problem!"
            )
            raise RuntimeError(msg) from ce
        self.components = {
            k: FlowchemComponentClient(v, parent=self)
            for k, v in self.device_info.components.items()
        }

    def __getitem__(self, item):
        """Get a device component by name or type."""
        if isinstance(item, type):
            return [
                component
                for component in self.components
                if isinstance(component, item)
            ]
        if isinstance(item, str):
            return self.components.get(item, None)
        return None

    # noinspection PyUnusedLocal
    @staticmethod
    def raise_for_status(resp, *args, **kwargs):  # noqa
        """Raise errors for status codes != 200 in session requests."""
        resp.raise_for_status()

    @staticmethod
    def log_responses(resp, *args, **kwargs):  # noqa
        """Log all the requests sent."""
        logger.debug(f"Reply: '{resp.text}' on {resp.url}")



"""This module is used to discover the serial address of any Vapourtec device connected to the PC."""
import asyncio
from textwrap import dedent

from loguru import logger

from flowchem.devices import R4Heater
from flowchem.utils.exceptions import InvalidConfigurationError


# noinspection PyProtectedMember
def r4_finder(serial_port) -> set[str]:
    """Try to initialize an R4Heater on every available COM port."""
    logger.debug(f"Looking for R4Heaters on {serial_port}...")
    # Static counter for device type across different serial ports
    if "counter" not in r4_finder.__dict__:
        r4_finder.counter = 0  # type: ignore

    try:
        r4 = R4Heater(port=serial_port)
    except InvalidConfigurationError as ic:
        logger.error("config - {}".format(ic.args[0]))
        return set()

    try:
        asyncio.run(r4.initialize())
    except InvalidConfigurationError:
        r4._serial.close()
        return set()

    if r4.device_info.version:
        logger.info(f"R4 version {r4.device_info.version} found on <{serial_port}>")
        # Local variable for enumeration
        r4_finder.counter += 1  # type: ignore
        cfg = f"[device.r4-heater-{r4_finder.counter}]"  # type:ignore
        cfg += dedent(
            f"""
        type = "R4Heater"
        port = "{serial_port}"\n\n""",
        )
    else:
        cfg = ""
    logger.info(f"Close the serial port: <{serial_port}>")
    r4._serial.close()
    return set(cfg)



"""Knauer DAD component."""
from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.sensors.photo_sensor import PhotoSensor

if TYPE_CHECKING:
    from flowchem.devices.knauer.dad import KnauerDAD
from flowchem.components.technical.power import PowerSwitch


class KnauerDADLampControl(PowerSwitch):
    """
    Control the lamp of the Knauer DAD device.

    Attributes:
        hw_device (KnauerDAD): The hardware device for the Knauer DAD.
    """
    hw_device: KnauerDAD

    def __init__(self, name: str, hw_device: KnauerDAD) -> None:
        """
        Initialize the KnauerDADLampControl component.

        Args:
            name (str): The name of the component.
            hw_device (KnauerDAD): The hardware device instance for controlling the Knauer DAD lamp.
        """
        super().__init__(name, hw_device)
        self.lamp = name
        self.add_api_route("/lamp_status", self.get_lamp, methods=["GET"])
        self.add_api_route("/status", self.get_status, methods=["GET"])

    async def get_status(self) -> str:
        """
        Get the status of the instrument.

        Returns:
            str: The status of the instrument.
        """
        return await self.hw_device.status()

    async def get_lamp(self):
        """
        Get the status of the lamp.

        Returns:
            str: The status of the lamp.
        """
        return await self.hw_device.lamp(self.lamp)
        # return {

    async def power_on(self):
        """
        Turn the lamp power on.

        Returns:
            str: The response from the hardware device.
        """
        return await self.hw_device.lamp(self.lamp, "ON")

    async def power_off(self):
        """
        Turn the lamp power off.

        Returns:
            str: The response from the hardware device.
        """
        return await self.hw_device.lamp(self.lamp, "OFF")

    async def set_lamp(self, state: str) -> str:
        """
        Set the lamp state.

        Args:
            state (str): The desired state of the lamp ("ON" or "OFF").

        Returns:
            str: The response from the hardware device.
        """
        return await self.hw_device.lamp(self.lamp, state)
        # match lamp_name:
        #     case "d2":
        #         await self.hw_device.d2(state)
        #     case "hal":
        #         await self.hw_device.hal(state)
        #     case _:
        #         logger.error("unknown lamp name!")


class DADChannelControl(PhotoSensor):
    """
    Control a specific channel of the Knauer DAD device.

    Attributes:
        hw_device (KnauerDAD): The hardware device for the Knauer DAD.
        channel (int): The channel number to control.
    """
    hw_device: KnauerDAD

    def __init__(self, name: str, hw_device: KnauerDAD, channel: int) -> None:
        """
        Initialize the DADChannelControl component.

        Args:
            name (str): The name of the component.
            hw_device (KnauerDAD): The hardware device instance for controlling the Knauer DAD channel.
            channel (int): The channel number to control.
        """
        super().__init__(name, hw_device)
        self.channel = channel

        # additional parameters
        self.add_api_route("/set-wavelength", self.set_wavelength, methods=["PUT"])
        self.add_api_route(
            "/set-integration-time",
            self.set_integration_time,
            methods=["PUT"],
        )
        self.add_api_route("/set-bandwidth", self.set_bandwidth, methods=["PUT"])

        # Ontology: diode array detector
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/CHMO_0002503",
        )

    async def acquire_signal(self) -> float:
        """
        Acquire a signal from the sensor.

        Returns:
            float: The acquired signal.
        """
        return await self.hw_device.read_signal(self.channel)

    async def set_wavelength(self, wavelength: int):
        """
        Set the acquisition wavelength.

        Be aware that wavelength=0 means that nothing will be collected.

        Args:
            wavelength (int): The desired wavelength in nm (0-999 nm).

        Returns:
            str: The response from the hardware device.
        """
        return await self.hw_device.set_wavelength(self.channel, wavelength)

    async def set_integration_time(self, int_time: int):
        """
        Set the integration time.

        Args:
            int_time (int): The desired integration time in ms (10 - 2000 ms).

        Returns:
            str: The response from the hardware device.
        """
        return await self.hw_device.integration_time(int_time)

    async def set_bandwidth(self, bandwidth: int):
        """
        Set the bandwidth.

        Args:
            bandwidth (int): The desired bandwidth in nm (4 to 25 nm).

        Returns:
            str: The response from the hardware device.
        """
        return await self.hw_device.bandwidth(bandwidth)

    async def set_shutter(self, status: str): # Todo: Not used
        """
        Set the shutter status.

        Args:
            status (str): The desired shutter status ("CLOSED", "OPEN", or "FILTER").

        Returns:
            str: The response from the hardware device.
        """
        return await self.hw_device.shutter(status)

    async def power_on(self) -> str:
        """
        Check the lamp status.

        Returns:
            str: The status of both the D2 and halogen lamps.
        """
        return f"d2 lamp is {await self.hw_device.lamp('d2')}; halogen lamp is {await self.hw_device.lamp('hal')}"

    async def power_off(self) -> str:
        """
        Deactivate the measurement channel.

        Returns:
            str: The response from the hardware device.
        """
        reply = await self.hw_device.set_wavelength(self.channel, 0)
        return reply



"""This module is used to discover the serial address of any ML600 connected to the PC."""
import asyncio
from textwrap import dedent

from loguru import logger

from flowchem.devices.hamilton.ml600 import HamiltonPumpIO, InvalidConfigurationError


def ml600_finder(serial_port) -> set[str]:
    """Try to initialize an ML600 on every available COM port."""
    logger.debug(f"Looking for ML600 pumps on {serial_port}...")
    # Static counter for device type across different serial ports
    if "counter" not in ml600_finder.__dict__:
        ml600_finder.counter = 0  # type: ignore
    dev_config: set[str] = set()

    try:
        link = HamiltonPumpIO.from_config({"port": serial_port})
    except InvalidConfigurationError:
        return dev_config

    try:
        asyncio.run(link.initialize(hw_initialization=False))
    except InvalidConfigurationError:
        # This is necessary only on failure to release the port for the other inspector
        link._serial.close()
        return dev_config

    for count in range(link.num_pump_connected):
        logger.info(f"Pump ML600 found on <{serial_port}> address {count + 1}")

        ml600_finder.counter += 1  # type: ignore
        dev_config.add(
            dedent(
                f"\n\n[device.ml600-{ml600_finder.counter}]"  # type: ignore
                f"""type = "ML600"
                port = "{serial_port}"
                address = {count + 1}
                syringe_volume = "XXX ml" # Specify syringe volume here!\n""",
            ),
        )
    logger.info(f"Close the serial port: <{serial_port}>")
    link._serial.close()
    return dev_config



"""An HPLC control component."""
from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.flowchem_component import FlowchemComponent

if TYPE_CHECKING:
    from flowchem.devices.flowchem_device import FlowchemDevice


class HPLCControl(FlowchemComponent):
    """
    A component for controlling an HPLC (High Performance Liquid Chromatography) system.

    Attributes:
    -----------
    hw_device : FlowchemDevice
        The hardware device (HPLC) this component interfaces with.

    Methods:
    --------
    send_method(method_name: str):
        Submit a method to the HPLC system. This is useful when automatic actions, like switching a valve, are triggered.
    run_sample(sample_name: str, method_name: str):
        Run an HPLC sample using the specified sample and method.
    """
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Constructs all the necessary attributes for the HPLCControl object.

        Parameters:
        -----------
        name : str
            The name of the HPLC control component.
        hw_device : FlowchemDevice
            The hardware device (HPLC) this component interfaces with.
        """
        super().__init__(name, hw_device)
        self.add_api_route("/run-sample", self.run_sample, methods=["PUT"])
        self.add_api_route("/send-method", self.send_method, methods=["PUT"])

        # Ontology: high performance liquid chromatography instrument
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/OBI_0001057",
        )
        self.component_info.type = "HPLC Control"

    async def send_method(self, method_name):
        """
        Set or load a analytical method to the HPLC system.

        This method is useful for scenarios where an injection is automatically triggered, such as when switching a valve.

        Note that the commands is wrapped in the run sample

        Parameters:
        -----------
        method_name : str
            The name of the method to be submitted.

        Returns:
        --------
        None
        """
        ...

    async def run_sample(self, sample_name: str, method_name: str):
        """
        Run an HPLC sample using the provided sample name and method.

        Parameters:
        -----------
        sample_name : str
            The name of the sample to run.
        method_name : str
            The name of the method to use for running the sample.

        Returns:
        --------
        None
        """
        ...



"""Control module for the Vapourtec R4 heater."""
from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.technical.temperature import TemperatureControl, TempRange

if TYPE_CHECKING:
    from .r4_heater import R4Heater


class R4HeaterChannelControl(TemperatureControl):
    """Control class for a single channel of the R4 reactor heater."""

    hw_device: R4Heater  # for typing's sake

    def __init__(
        self,
        name: str,
        hw_device: R4Heater,
        channel: int,
        temp_limits: TempRange,
    ) -> None:
        """
        Initialize the R4HeaterChannelControl with a name, hardware device, channel, and temperature limits.

        Args:
            name (str): The name of the heater channel control.
            hw_device (R4Heater): The R4 heater hardware device.
            channel (int): The channel number to control.
            temp_limits (TempRange): The temperature limits for this channel.
        """
        super().__init__(name, hw_device, temp_limits)
        self.channel = channel

    async def set_temperature(self, temp: str):
        """
        Set the target temperature for this channel using a "magnitude and unit" format string.

        Args:
            temp (str): The desired temperature as a string (e.g., '50 °C', '75.5 °C').

        Returns:
            Awaitable: Result of the set temperature operation from the hardware device.
        """
        set_t = await super().set_temperature(temp)
        return await self.hw_device.set_temperature(self.channel, set_t)

    async def get_temperature(self) -> float:  # type: ignore
        """
        Retrieve the current temperature of this channel in degrees Celsius.

        Returns:
            float: The current temperature in degrees Celsius.
        """
        return float(await self.hw_device.get_temperature(self.channel))

    async def is_target_reached(self) -> bool:  # type: ignore
        """
        Check if the set temperature target has been reached for this channel.

        Returns:
            bool: True if the target temperature has been reached, False otherwise.
        """
        status = await self.hw_device.get_status(self.channel)
        return status.state == "S"

    async def power_on(self):
        """
        Turn on the temperature control for this channel.

        Returns:
            Awaitable: Result of the power on operation from the hardware device.
        """
        return await self.hw_device.power_on(self.channel)

    async def power_off(self):
        """
        Turn off the temperature control for this channel.

        Returns:
            Awaitable: Result of the power off operation from the hardware device.
        """
        return await self.hw_device.power_off(self.channel)



"""Functions for the construction of XML requests for Spinsolve."""
from pathlib import WindowsPath

from lxml import etree


def create_message(sub_element_name, attributes=None):
    """Create a minimal XML tree with Message as root and sub_element as child tag."""
    if attributes is None:
        attributes = {}

    root = etree.Element("Message")
    etree.SubElement(root, sub_element_name, attributes)
    return root


def set_attribute(name, value="") -> etree._Element:
    """Create a Set <Message>.

    Used for name = {Solvent | Sample} + indirectly by UserData and DataFolder.
    """
    base = create_message("Set")
    attribute = etree.SubElement(base.find("./Set"), name)
    attribute.text = value
    return base


def get_request(name) -> etree._Element:
    """Create a Get <Message> element.

    Used for name = {Solvent | Sample | UserData} + indirectly by UserData and DataFolder.
    """
    base = create_message("GetRequest")
    etree.SubElement(base.find("./GetRequest"), name)
    return base


def set_data_folder(location) -> etree._Element:
    """Create a Set DataFolder message."""
    # Get base request
    data_folder = set_attribute("DataFolder")

    # Add folder specific tag
    data_folder_node = data_folder.find(".//DataFolder")
    if data_folder_node:
        full_tree = etree.SubElement(data_folder, "TimeStampTree")
        full_tree.text = (
            location.as_posix() if isinstance(location, WindowsPath) else location
        )

    return data_folder


def create_protocol_message(name: str, options: dict) -> etree._Element:
    """Create an XML request to run a protocol."""
    xml_root = create_message("Start", {"protocol": name})
    start_tag = xml_root.find("Start")

    # All options are sent as Start tag SubElements
    for key, value in options.items():
        etree.SubElement(start_tag, "Option", {"name": f"{key}", "value": f"{value}"})

    return xml_root



"""A Diode Array Detector control component."""


# class DADSwitch(FlowchemComponent):
#     def __init__(self, name: str, hw_device: FlowchemDevice):
#         """DAD Control component."""
#         super().__init__(name, hw_device)
#         self.add_api_route("/lamp", self.get_lamp, methods=["GET"])
#         self.add_api_route("/lamp", self.set_lamp, methods=["PUT"])
#

#
#     async def get_lamp(self):
#         """Lamp status."""
#         ...
#
#     async def set_lamp(self, state: bool):
#         """Lamp status."""
#         ...
#
#
# class DADControl(FlowchemComponent):
#     def __init__(self, name: str, hw_device: FlowchemDevice):
#         """NMR Control component."""
#         super().__init__(name, hw_device)
#         self.add_api_route()
#         self.add_api_route("/acquire-spectrum", self.acquire_signal, methods=["PUT"])
#         self.add_api_route("/stop", self.stop, methods=["PUT"])
#
#
#
#     async def acquire_signal(self):
#         """Acquire an ."""
#         ...
#
#     async def stop(self):
#         """Stops acquisition and exit gracefully."""
#         ...



"""Azura compact sensor component."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .azura_compact import AzuraCompact

from flowchem.components.sensors.pressure_sensor import PressureSensor


class AzuraCompactSensor(PressureSensor):
    hw_device: AzuraCompact  # for typing's sake

    async def read_pressure(self, units: str = "bar"):
        """Read from sensor, result to be expressed in units (optional)."""
        pressure = await self.hw_device.read_pressure()
        return pressure.m_as(units)



"""This module is used to discover the serial address of any Elite11 connected to the PC."""
import asyncio
from textwrap import dedent

from loguru import logger

from flowchem.devices.harvardapparatus.elite11 import Elite11, HarvardApparatusPumpIO
from flowchem.utils.exceptions import InvalidConfigurationError


# noinspection PyProtectedMember
def elite11_finder(serial_port) -> set[str]:
    """Try to initialize an Elite11 on every available COM port. [Does not support daisy-chained Elite11!]."""
    logger.debug(f"Looking for Elite11 pumps on {serial_port}...")
    # Static counter for device type across different serial ports
    if "counter" not in elite11_finder.__dict__:
        elite11_finder.counter = 0  # type: ignore
    cfg: set[str] = set()

    try:
        link = HarvardApparatusPumpIO(port=serial_port)
    except InvalidConfigurationError:
        # This is necessary only on failure to release the port for the other inspector
        return cfg

    # Check for echo
    link._serial.write(b"\r\n")
    if link._serial.readline() != b"\n":
        # This is necessary only on failure to release the port for the other inspector
        link._serial.close()
        return cfg

    # Parse status prompt
    pump = link._serial.readline().decode("ascii")
    address = int(pump[0:2]) if pump[0:2].isdigit() else 0

    try:
        test_pump = Elite11(
            link,
            syringe_diameter="20 mm",
            syringe_volume="10 ml",
            address=address,
        )
        asyncio.run(test_pump.pump_info())
    except InvalidConfigurationError:
        # This is necessary only on failure to release the port for the other inspector
        link._serial.close()
        return cfg

    logger.info(f"Elite11 found on <{serial_port}>")

    # Local variable for enumeration
    elite11_finder.counter += 1  # type: ignore
    msg = f"[device.elite11-{elite11_finder.counter}]"  # type:ignore
    msg += dedent(
        f"""
                   type = "Elite11"
                   port = "{serial_port}"
                   address = {address}
                   syringe_diameter = "XXX mm" # Specify syringe diameter!
                   syringe_volume = "YYY ml" # Specify syringe volume!\n\n""",
    )
    logger.info(f"Close the serial port: <{serial_port}>")
    link._serial.close()
    cfg.add(msg)
    return cfg


from dataclasses import dataclass

from flowchem import ureg


@dataclass
class PBCommand:
    """Class representing a PBCommand."""

    command: str

    def to_chiller(self) -> bytes:
        """Validate and encode to bytes array to be transmitted."""
        self.validate()
        return self.command.encode("ascii")

    def validate(self):
        """Check command structure to be compliant with PB format."""
        if len(self.command) == 8:
            self.command += "\r\n"
        # 10 characters
        assert len(self.command) == 10
        # Starts with {
        assert self.command[0] == "{"
        # M for master (commands) S for slave (replies).
        assert self.command[1] in ("M", "S")
        # Address, i.e. the desired function. Hex encoded.
        assert 0 <= int(self.command[2:4], 16) < 256
        # Value
        assert self.command[4:8] == "****" or 0 <= int(self.command[4:8], 16) <= 65536
        # EOL
        assert self.command[8:10] == "\r\n"

    @property
    def data(self) -> str:
        """Data portion of PBCommand."""
        return self.command[4:8]

    def parse_temperature(self) -> float:
        """Parse a device temp from hex string to Celsius float."""
        # self.data is the two's complement 16-bit signed hex, see manual
        temp = (
            (int(self.data, 16) - 65536) / 100
            if int(self.data, 16) > 32767
            else (int(self.data, 16)) / 100
        )
        # Note: -151 used for invalid temperatures!
        return temp

    def parse_integer(self) -> int:
        """Parse a device reply from hexadecimal string to base 10 integers."""
        return int(self.data, 16)

    def parse_rpm(self) -> str:
        """Parse a device reply from hexadecimal string to rpm."""
        return str(ureg.Quantity(f"{self.parse_integer()} rpm"))

    def parse_bits(self) -> list[bool]:
        """Parse a device reply from hexadecimal string to 16 constituting bits."""
        bits = f"{int(self.data, 16):016b}"
        return [bool(int(x)) for x in bits]

    def parse_boolean(self):
        """Parse a device reply from hexadecimal string (0x0000 or 0x0001) to boolean."""
        return self.parse_integer() == 1

    def parse_status1(self) -> dict[str, bool]:
        """Parse response to status1 command and returns dict."""
        bits = self.parse_bits()
        return {
            "temp_ctl_is_process": bits[0],
            "circulation_active": bits[1],
            "refrigerator_on": bits[2],
            "temp_is_process": bits[3],
            "circulating_pump": bits[4],
            "cooling_power_available": bits[5],
            "tkeylock": bits[6],
            "is_pid_auto": bits[7],
            "error": bits[8],
            "warning": bits[9],
            "int_temp_mode": bits[10],
            "ext_temp_mode": bits[11],
            "dv_e_grade": bits[12],
            "power_failure": bits[13],
            "freeze_protection": bits[14],
        }

    def parse_status2(self) -> dict[str, bool]:
        """Parse response to status2 command and returns dict. See manufacturer docs for more info."""
        bits = self.parse_bits()
        return {
            "controller_is_external": bits[0],
            "drip_tray_full": bits[5],
            "venting_active": bits[7],
            "venting_successful": bits[8],
            "venting_monitored": bits[9],
        }



"""Generic valve."""
from __future__ import annotations

from pydantic import BaseModel
import json
from typing import Tuple

from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.utils.exceptions import InvalidConfigurationError, DeviceError


def return_tuple_from_input(str_or_tuple):
    # in case no input is given, required, simply return None, will be dealt with by consumer
    if not str_or_tuple:
        return None
    elif type(str_or_tuple) is str:
        parsed_input = json.loads(str_or_tuple)
        if type(parsed_input[0]) is not list:
            parsed_input = [parsed_input]
        return tuple(tuple(inner) for inner in parsed_input)
    elif type(str_or_tuple) is tuple:
        if type(str_or_tuple[0]) is not tuple:
            str_or_tuple = (str_or_tuple,)
        return str_or_tuple
    else:
        raise DeviceError("Please provide input of type '[[1,2],[3,4]]'")


def return_bool_from_input(str_or_bool):
    if type(str_or_bool) is bool:
        return str_or_bool
    elif type(str_or_bool) is str:
        if str_or_bool.lower() == "true":
            return True
        elif str_or_bool.lower() == "false":
            return False
        elif str_or_bool == "":
            return None
        else:
            raise DeviceError("Please provide input of type bool, '' or 'True' or 'False'")


class ValveInfo(BaseModel):
    """
    ports: an attribute representing the available ports on the stator
    positions: an attribute mapping implicit, tacit numbers as keys to the stator ports that are connected at this
                position
    """
    ports: list[tuple]
    positions: dict[int, Tuple[Tuple[None | int, ...], ...]]


# mypy: ignore-errors
def all_tuples_in_nested_tuple(tuple_in: Tuple[Tuple[int, int], ...],
                               tuple_contains: Tuple[Tuple[int, int, ...], ...]) -> bool:
    """Check if all requested tuples are in a tuple of tuples"""
    all_contained = []
    for subtuple in tuple_in:
        for supertuple in tuple_contains:
            if set(subtuple) <= set(supertuple):
                all_contained.append(True)
                break
    if len(all_contained) == len(tuple_in):
        return True
    else:
        return False


# mypy: ignore-errors
def no_tuple_in_nested_tuple(tuple_in: Tuple[Tuple[int, int], ...],
                             tuple_contains: Tuple[Tuple[int, int, ...], ...]) -> bool:
    """Check if none of requested tuples are in a tuple of tuples"""
    contains_tuple = False
    for subtuple in tuple_in:
        for supertuple in tuple_contains:
            if set(subtuple) <= set(supertuple):
                contains_tuple = True
    return not contains_tuple


class Valve(FlowchemComponent):
    """An abstract class for devices of type valve.

    ... warning::
        Device objects should not directly generate components with this object but rather a more specific valve type,
        such as `SixPortTwoPositionValve` or `SixPortPositionValve`.

    All valves are characterized by:

    - a connections() method, which returns an Instance of the ValveInfo class
    - a `set_position()` method
    - a `get_position()` method

    This is explicit and informative in itself and requires no further intermittant helper mappings
    """

    def __init__(
            self,
            name: str,
            hw_device: "FlowchemDevice",
            stator_ports: [(), ()], # type: ignore
            rotor_ports: [(), ()], # type: ignore
    ) -> None:
        """Create a valve object.

        Args:
        ----
            name: device name, passed to FlowchemComponent.
            hw_device: the object that controls the hardware.
            stator_ports
            rotor_ports
            rotor and stator are both represented like:
            (   (1,2,3,4,5,6),          (0))
                radial ports       middle ports
                          should be equally distributed, with equally spaced angle in between. If this is not the
                          case, add None
             for missing port

            Exemple: 4 port 5 Position Valve
            stator_ports=[(None, None, 1, None, 2, None, 3, None,), (0,)],
            rotor_ports=[(None, 5, None, None, 4, None, 4, None), (5,)],

                                 1
                             * * * * *
                           *     x     *
                         *               *
                     8  *  x            -x *  2
                        *            -     *
                     7  * o-      o-     o *  3
                        *   -              *
                        *  x -           x *  4
                     6   *     -          *
                           *    -o      *
                             * * * * *
                                 5

        X - inaccessible port (addressed as None in stator_ports)
        O - accessible port

        Note that the only positions available as accessible ports are 3, 5, 7, and 0 (corresponding to 1, 2, 3, and 0).

        The connection is made between the port at position 5 and the ports at positions 7 and 0 (central port) and
        at position 2.

        """
        # a valve consists of a rotor and a stator. Solenoid valves Are special cases and can be decomposed into
        # Open/closed valves, need not be treated here but could be simulated by a [1,2,None] and rotor [3,3,None]
        self._rotor_ports = rotor_ports
        self._stator_ports = stator_ports
        self._positions = self._create_connections(self._stator_ports, self._rotor_ports)

        # bwe can infer
        super().__init__(name, hw_device)

        self.add_api_route("/position", self.get_position, methods=["GET"])
        self.add_api_route("/position", self.set_position, methods=["PUT"])
        self.add_api_route("/connections", self.connections, methods=["GET"])

    @staticmethod
    def _create_connections(stator_ports, rotor_ports):
        """
        Create possible switching states from a stator and rotor representation. Position names are integers. Going to
        the next position in clockwise direction increases position name by one
        """
        connections = {}
        if len(rotor_ports) != len(stator_ports):
            raise InvalidConfigurationError
        if len(rotor_ports) == 1:
            # in case there is no 0 port, for data uniformity, internally add it. strictly, the stator and rotor
            # should reflect physical properties, so if stator has a hole in middle it should have 0, but only rotor
            # None. Since this does not impact functionality, thoroughness will be left to the user
            rotor_ports.append([None])
            stator_ports.append([None])
        # it is rather simple: we just move the rotor by one and thereby create a dictionary
        for _ in range(len(rotor_ports[0])):
            rotor_curr = rotor_ports[0][-_:] + rotor_ports[0][:-_]
            _connections_per_position = {}
            for rotor_position, stator_position in zip(rotor_curr + rotor_ports[1], stator_ports[0] + stator_ports[1]):
                # rotor positions act as dictionary keys, take into account the [1] position for connecting the 0
                # if dict key exists, instead of overwriting, simply append
                # if rotor is none, means there is no connection, so do not add
                if rotor_position is not None:
                    try:
                        _connections_per_position[rotor_position] += (stator_position,)
                    except KeyError:
                        _connections_per_position[rotor_position] = (stator_position,)
                        # get rid of the keys, values are the connected ports in each position
            connections[_] = tuple(_connections_per_position.values())
        # lastly, trim the lists of connections that already exist
        # trim returned connections
        unique_connections = set(connections.values())
        to_delete = []
        for connection in unique_connections:
            counter = 0
            for key, value in connections.items():
                if connection == value:
                    if counter > 0:
                        to_delete.append(key)
                    counter += 1
        for _ in to_delete:
            del connections[_]

        return connections

    def _change_connections(self, raw_position: int | str, reverse: bool = False):
        # abstract valve mapping needs to be translated to device-specific position naming. This can be e.g.
        # addition/subtraction of one, multiplication with some angle or mapping to letters. Needs to be implemented on
        # device level since this is device communication protocol specific
        raise NotImplementedError

    def _connect_positions(self,
                           positions_to_connect: Tuple[Tuple[int, int], ...],
                           positions_not_to_connect: Tuple[Tuple[int, int], ...] | None = None,
                           arbitrary_switching: bool = True) -> int:
        """
        This is the heart of valve switching logic: select the suitable position (so actually the key in
        self._positions) to create desired connections
        """
        possible_positions = []
        # check if this is possible given the mapping
        for key, values in self._positions.items():
            if positions_not_to_connect:
                if all_tuples_in_nested_tuple(positions_to_connect, values) and no_tuple_in_nested_tuple(
                        positions_not_to_connect, values):
                    possible_positions.append(key)
            elif all_tuples_in_nested_tuple(positions_to_connect, values):
                possible_positions.append(key)
        if len(possible_positions) > 1:
            if not arbitrary_switching:
                raise DeviceError("There are multiple positions for the valve to connect your specified ports. "
                                  "Either allow arbitrary switching, or specify which connections not to connect")
            elif arbitrary_switching:
                return possible_positions[0]
            else:
                raise DeviceError
        elif len(possible_positions) == 1:
            return possible_positions[0]
        else:
            # this means length == 0, no connection possible
            raise DeviceError("Connection is not possible. The valve you selected can not connect selected ports."
                              "This can be due to exclusion of certain connections by setting positions_not_to_connect")

    async def get_position(self) -> list[list[int]]:
        """Get current valve position."""
        if not hasattr(self, "identifier"):
            pos = await self.hw_device.get_raw_position() # type: ignore
        else:
            pos = await self.hw_device.get_raw_position(self.identifier) # type: ignore
        pos = int(pos) if pos.isnumeric() else pos
        return self._positions[int(self._change_connections(pos, reverse=True))]

    async def set_position(self,
                           connect: str = "",
                           disconnect: str = "",
                           ambiguous_switching: str | bool = False):
        """Move valve to position, which connects named ports"""
        connect_tuple = return_tuple_from_input(connect)
        disconnect_tuple = return_tuple_from_input(disconnect)
        ambiguous_switching_bool = return_bool_from_input(ambiguous_switching)
        target_pos = self._connect_positions(positions_to_connect=connect_tuple,
                                             positions_not_to_connect=disconnect_tuple,
                                             arbitrary_switching=ambiguous_switching_bool)
        target_pos = self._change_connections(target_pos)
        if not hasattr(self, "identifier"):
            await self.hw_device.set_raw_position(target_pos) # type: ignore
        else:
            await self.hw_device.set_raw_position(target_pos, target_component=self.identifier) # type: ignore

    def connections(self) -> ValveInfo:
        """Get the list of all available positions for this valve.
        This mainly has informative purpose
        """
        return ValveInfo(ports=self._stator_ports, positions=self._positions)

    # Philosophy: explicitly specify which ports to connect
    # In case of a simple multi-position valve, it always connects the always open central port to the requested port.
    # But there are more complex rotors available, already with an injection valve: connect((1, 2)) is concise and
    # clear.
    # What is even more important, one can specify which ports to not connect (optionally during switching)
    # This issue is most pressing with hamilton valves, where some positions connect 3 ports, and it is very hard to
    # foresee what command does what. SO here a simple connect((1,2)) helps.
    # In order for that to work, and to make the coding and usage simple and concise, some definitions are needed:
    # 1) The port zero can exist, but does not necessarily.
    #    For nomenclature reasons, port zero is the one the turning axis and only this one. Commonly, this port,
    #    if existing, is always open
    # 2) At the physical valve, the upmost is port 1
    #   a) If there is no port straight on top, then one goes in clockwise direction, until a port comes, which is then one
    # 3 )Beware: For logical reasons, we need to introduce ports of "number" None. These are needed because we need to
    #   define dead-ends. These dead-ends are IMMUTABLE dead-ends, so the stator or rotor do not have an opening there
    #   Any time there is a different amount of positions on rotor and stator, None ports are introduced
    # 5) Mutable dead-ends: blanking plugs are treated as port number, the consumer needs to deal with its definition by
    #   graph or similar since blanking plugs on valve side could be open
    # Dead-ends are needed because we represent valves as graphs, edges are represented by same numbers shared. If a
    # port does not connect to anything, we set it None. There is 1 example where that is strictly needed for the logic
    # to work: Again the hamilton, it will become clear why. So much now: The rotor has more open positions than the
    # stator.
    # 6) The so far mentioned logic only strictly applies to valves facing the user with their front side, however,
    # e.g. the auto-sampler faces one valve with its always open port to the ground. Simply flip horizontally until it
    # faces you



"""Azura compact pump component."""
from __future__ import annotations

from typing import TYPE_CHECKING

from loguru import logger

from flowchem import ureg

if TYPE_CHECKING:
    from .azura_compact import AzuraCompact
from flowchem.components.pumps.hplc_pump import HPLCPump


def isfloat(rate: str) -> bool:
    try:
        float(rate)
        return True
    except ValueError:
        return False


class AzuraCompactPump(HPLCPump):
    """
    Azura Compact Pump component for interfacing with the AzuraCompact hardware.

    This class inherits from HPLCPump and provides specific implementations
    for controlling the Azura Compact pump device. It allows for starting,
    stopping, and querying the pump's status.

    Attributes:
        hw_device (AzuraCompact): An instance of the AzuraCompact hardware device.

    Methods:
        infuse(rate: str = "", volume: str = "") -> bool:
            Starts the infusion process with a specified flow rate. The volume parameter is ignored.
            If the rate is numeric without units, it assumes "ml/min" as the default unit.

        stop() -> None:
            Stops the infusion process.

        is_pumping() -> bool:
            Checks whether the pump is currently running.

    Parameters:
        name (str): The name of the pump.
        hw_device (AzuraCompact): An instance of the AzuraCompact hardware device.
    """

    hw_device: AzuraCompact  # for typing's sake

    def __init__(self, name: str, hw_device: AzuraCompact) -> None:
        """Initialize the AzuraCompactPump component.

        Args:
            name (str): The name of the pump.
            hw_device (AzuraCompact): An instance of the AzuraCompact hardware device.
        """
        super().__init__(name, hw_device)

    async def infuse(self, rate: str = "", volume: str = "") -> bool:
        """Start infusion with the specified flow rate.

        Args:
            rate (str): The desired flow rate, which can include units (e.g., "10 ml/min").
                        If the rate is a numeric value, it assumes "ml/min" as the unit.
            volume (str): The desired volume for infusion. This parameter is currently ignored
                          by the Azura Compact pump.

        Returns:
            bool: True if the infusion command was successfully sent, False otherwise.
        """
        if volume:
            logger.warning(f"Volume parameter ignored: not supported by {self.name}!")

        if isfloat(rate):
            rate = "0 ml/min"
        if rate.isnumeric():
            rate += " ml/min"
            logger.warning("Units missing, assuming ml/min!")

        parsed_rate = ureg.Quantity(rate)

        await self.hw_device.set_flow_rate(rate=parsed_rate)
        return await self.hw_device.infuse()

    async def stop(self):
        """Stop the pumping process.

        Stops the pump if it is currently running.
        """
        await self.hw_device.stop()

    async def is_pumping(self) -> bool:
        """Check if the pump is currently running.

        Returns:
            bool: True if the pump is running, False otherwise.
        """
        return self.hw_device.is_running()



"""Controls a local ClarityChrom instance via the CLI interface."""
# See https://www.dataapex.com/documentation/Content/Help/110-technical-specifications/110.020-command-line-parameters/110.020-command-line-parameters.htm
import asyncio
from pathlib import Path
from shutil import which

from loguru import logger

from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.utils.people import jakob, wei_hsin

from .clarity_hplc_control import ClarityComponent


class Clarity(FlowchemDevice):
    """
    Controls a local ClarityChrom instance via the CLI interface.

    This class manages software startup, configuration, and command execution for
    ClarityChrom chromatography software. It validates executable paths, constructs
    initialization commands, and handles CLI interactions with timeout management.

    Attributes:
    -----------
    executable : str
        Path to the ClarityChrom executable. Automatically quoted if spaces are detected.
    instrument_number : int
        Target instrument number for multi-instrument setups (default: 1).
    startup_time : float
        Time (seconds) allowed for software initialization before operation.
    cmd_timeout : float
        Maximum duration (seconds) allowed for individual command execution.
    _init_command : str
        Pre-built initialization command combining config file, credentials,
        and startup method parameters.

    Methods:
    --------
    initialize() -> None
        Start ClarityChrom with configured parameters and register components.
    execute_command(command: str, without_instrument_num: bool = False) -> bool
        Execute CLI commands with optional instrument number bypass.
    """
    def __init__(
        self,
        name,
        executable: str = r"C:\claritychrom\bin\claritychrom.exe",
        instrument_number: int = 1,
        startup_time: float = 20,
        startup_method: str = "",
        cmd_timeout: float = 3,
        user: str = "admin",
        password: str = "",
        cfg_file: str = "",
    ) -> None:
        """
        Constructs all the necessary attributes for the Clarity object.

        Parameters:
        -----------
        name : str
            The name of the Clarity instance.
        executable : str, optional
            Path to the ClarityChrom executable (default is r"C:\claritychrom\bin\claritychrom.exe").
        instrument_number : int, optional
            The instrument number to control (default is 1).
        startup_time : float, optional
            The time to wait for ClarityChrom to start up and become responsive (default is 20 seconds).
        startup_method : str, optional
            The startup method to use (default is an empty string).
        cmd_timeout : float, optional
            The timeout duration for command execution (default is 3 seconds).
        user : str, optional
            The username for ClarityChrom (default is "admin").
        password : str, optional
            The password for ClarityChrom (default is an empty string).
        cfg_file : str, optional (PATH\FILENAME)
            The configuration file for ClarityChrom (default is an empty string).
        """
        super().__init__(name=name)
        # Metadata
        self.device_info.authors = [jakob, wei_hsin]
        self.device_info.manufacturer = "DataApex"
        self.device_info.model = "Clarity Chromatography"

        # Validate executable
        if which(executable):
            self.executable = executable
        else:
            assert '"' not in executable
            self.executable = f'"{executable}"'
        assert which(executable) or Path(executable).is_file(), "Valid executable found"

        # Save instance variables
        self.instrument = instrument_number
        self.startup_time = startup_time
        self.cmd_timeout = cmd_timeout

        # Pre-form initialization command to avoid passing tons of vars to initialize()
        self._init_command = ""
        self._init_command += f" cfg={cfg_file}" if cfg_file else ""
        self._init_command += f" u={user}" if user else ""
        self._init_command += f" p={password}" if password else ""
        self._init_command += f' "{startup_method}"'

    async def initialize(self):
        """Start ClarityChrom and wait for it to be responsive."""
        await self.execute_command(self._init_command)
        logger.info(f"Clarity startup: waiting {self.startup_time} seconds")
        await asyncio.sleep(self.startup_time)
        self.components.append(ClarityComponent(name="clarity", hw_device=self))

    async def execute_command(self, command: str, without_instrument_num: bool = False):
        """
        Execute ClarityChrom CLI command with timeout handling.

        Commands in string format that are accepted by the device.
        There is a list of the command available.
        (See more detail in the documentation and/or the manual reference)

        Parameters:
        -----------
        command : str
            Command string to execute (without instrument specification).
        without_instrument_num : bool
            Skip adding instrument number parameter (for global commands).

        Returns:
        --------
        bool
            True if command completed successfully, False on timeout.
        """
        if without_instrument_num:
            cmd_string = self.executable + f" {command}"
        else:
            cmd_string = self.executable + f" i={self.instrument} {command}"

        logger.debug(f"Executing Clarity command `{command}`")
        process = await asyncio.create_subprocess_shell(cmd_string)
        try:
            await asyncio.wait_for(process.wait(), timeout=self.cmd_timeout)
            return True
        except TimeoutError:
            logger.error(f"Subprocess timeout expired (timeout = {self.cmd_timeout} s)")
            return False



import ipaddress

from loguru import logger
from pydantic import AnyHttpUrl
from zeroconf import ServiceInfo, ServiceListener, Zeroconf

from flowchem.client.device_client import FlowchemDeviceClient

FLOWCHEM_SUFFIX = "._labthing._tcp.local."
FLOWCHEM_TYPE = FLOWCHEM_SUFFIX[1:]


def zeroconf_name_to_device_name(zeroconf_name: str) -> str:
    assert zeroconf_name.endswith(FLOWCHEM_SUFFIX)
    return zeroconf_name[: -len(FLOWCHEM_SUFFIX)]


def flowchem_devices_from_url_dict(
    url_dict: dict[str, AnyHttpUrl],
) -> dict[str, FlowchemDeviceClient]:
    dev_dict = {}
    for name, url in url_dict.items():
        dev_dict[name] = FlowchemDeviceClient(url)
    return dev_dict


def device_url_from_service_info(
    service_info: ServiceInfo,
    device_name: str,
    active_ip: list | None = None
) -> AnyHttpUrl | None:

    if not active_ip:
        if service_info.addresses:
            # Needed to convert IP from bytes to str
            device_ip = ipaddress.ip_address(service_info.addresses[0])
            return AnyHttpUrl(f"http://{device_ip}:{service_info.port}/{device_name}")

        logger.warning(f"No address found for {device_name}!")
        return None

    else:
        device_ip = ipaddress.ip_address(active_ip[0])
        return AnyHttpUrl(f"http://{device_ip}:{service_info.port}/{device_name}")


class FlowchemCommonDeviceListener(ServiceListener):
    def __init__(self, response_ips=False) -> None:
        self.flowchem_devices: dict[str, AnyHttpUrl] = {}
        if response_ips:
            from zeroconf._utils.net import create_sockets
            listen_socket, respond_sockets = create_sockets()
            self.active_ips = []
            for socket in respond_sockets:
                self.active_ips.append(socket.getsockname()[0])
            print(self.active_ips)
        else:
            self.active_ips = []

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        logger.debug(f"Service {zeroconf_name_to_device_name(name)} removed")
        self.flowchem_devices.pop(name, None)

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        logger.debug(f"Service {zeroconf_name_to_device_name(name)} updated")
        self.flowchem_devices.pop(name, None)
        self._save_device_info(zc, type_, name, self.active_ips)  #todo: need?

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        logger.debug(f"Service {zeroconf_name_to_device_name(name)} added")
        self._save_device_info(zc, type_, name, self.active_ips)  #todo: need?

    def _save_device_info(self, zc: Zeroconf, type_: str, name: str, active_ips=None) -> None:
        raise NotImplementedError



"""Syringe pump component, two flavours, infuse only, infuse-withdraw."""
from flowchem.components.pumps.pump import Pump
from flowchem.devices.flowchem_device import FlowchemDevice


class SyringePump(Pump):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(name, hw_device)

        # Ontology: Syringe pump
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/OBI_0400100",
        )
        self.component_info.type = "Syringe Pump"



"""Pressure control."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pint
from loguru import logger

from flowchem import ureg
from flowchem.components.flowchem_component import FlowchemComponent

if TYPE_CHECKING:
    from flowchem.devices.flowchem_device import FlowchemDevice


class PressureControl(FlowchemComponent):
    """
    A generic pressure controller for managing and monitoring pressure.

    This component provides an interface to set and get pressure values, check if the target pressure has been reached,
    and control the power state of the pressure system.

    Attributes:
    -----------
    hw_device : FlowchemDevice
        The hardware device instance associated with this pressure control component.

    Methods:
    --------
    set_pressure(pressure: str) -> pint.Quantity:
        Set the target pressure using a string representation.
    get_pressure() -> float:
        Retrieve the current pressure from the device.
    is_target_reached() -> bool:
        Check if the target pressure has been reached.
    power_on() -> bool:
        Turn on the pressure control.
    power_off() -> bool:
        Turn off the pressure control.
    """

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Initialize the PressureControl component.

        Parameters:
        -----------
        name : str
            The name assigned to this pressure control component.
        hw_device : FlowchemDevice
            The hardware device instance associated with this pressure control component.
        """
        super().__init__(name, hw_device)

        self.add_api_route("/pressure", self.set_pressure, methods=["PUT"])
        self.add_api_route("/pressure", self.get_pressure, methods=["GET"])

        self.add_api_route("/power-on", self.power_on, methods=["PUT"])
        self.add_api_route("/power-off", self.power_off, methods=["PUT"])

        self.add_api_route("/target-reached", self.is_target_reached, methods=["GET"])

    async def set_pressure(self, pressure: str) -> pint.Quantity:
        """
        Set the target pressure using a string representation.

        This method interprets the provided string to set the pressure. If no units are specified,
        'mbar' is assumed as the default unit.

        Parameters:
        -----------
        pressure : str
            The target pressure to be set, expressed in a string format.

        Returns:
        --------
        pint.Quantity
            The pressure value as a `pint.Quantity` object with units of 'mbar'.
        """
        # Add units (mbar) if none
        try:
            float(pressure)
        except ValueError:
            pass
        else:
            logger.warning("No units provided to set_pressure, assuming mbar.")
            pressure = pressure + "mbar"
        return ureg.Quantity(pressure)

    async def get_pressure(self) -> float:  # type: ignore
        """
        Retrieve the current pressure from the device.

        This method should be overridden in a subclass to interact with the specific hardware.

        Returns:
        --------
        float
            The current pressure.
        """
        ...

    async def is_target_reached(self) -> bool:  # type: ignore
        """
        Check if the target pressure has been reached.

        This method should be overridden in a subclass to provide the actual implementation
        for checking the pressure status.

        Returns:
        --------
        bool
            Returns True if the target pressure has been reached, False otherwise.
        """
        ...

    async def power_on(self):
        """
        Turn on the pressure control.

        This method should be implemented in a subclass to send the appropriate command
        to the hardware to activate the pressure control.

        Returns:
        --------
        str
            Returns binary string if the operation was successful.
        """
        ...

    async def power_off(self):
        """
        Turn off the pressure control.

        This method should be implemented in a subclass to send the appropriate command
        to the hardware to deactivate the pressure control.

        Returns:
        --------
        str
            Returns binary string if the operation was successful.
        """
        ...



"""Entry-point module for the command line prefixes, called in case you use `python -m flowchem`.
Why does this file exist, and why `__main__`? For more info, read:
- https://www.python.org/dev/peps/pep-0338/
- https://docs.python.org/3/using/cmdline.html#cmdoption-m.
"""
import asyncio
import sys
from pathlib import Path

import rich_click as click
import uvicorn
from loguru import logger

from flowchem import __version__
from flowchem.server.core import Flowchem


@click.argument("device_config_file", type=click.Path(), required=True)
@click.option(
    "-l",
    "--log",
    "logfile",
    type=click.Path(),
    default=None,
    help="Save logs to file.",
)
@click.option(
    "-h",
    "--host",
    "host",
    type=str,
    default="0.0.0.0",
    help="Server host. 0.0.0.0 is used to bind to all addresses, do not use for internet-exposed devices!",
)
@click.option("-d", "--debug", is_flag=True, help="Print debug info.")
@click.version_option()
@click.command()
def main(device_config_file, logfile, host, debug):
    """Flowchem main program.

    Parse device_config_file and starts a server exposing the devices via REST-ful API.

    Args:
    ----
        device_config_file: Flowchem configuration file specifying device connection settings (TOML)
        logfile: Output file for logs.
        host: IP on which the server will be listening. Loopback IP as default, use LAN IP to enable remote access.
        debug: Print debug info
    """
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    if not debug:
        # Set stderr to info
        logger.remove()
        logger.add(sys.stderr, level="INFO")

    logger.info(f"Starting flowchem v. {__version__}!")
    if logfile:
        logger.add(Path(logfile), level="DEBUG")
    logger.debug(f"Starting server with configuration file: '{device_config_file}'")

    async def main_loop():
        """Main application loop, the event loop is shared between uvicorn and flowchem."""
        flowchem = Flowchem()
        await flowchem.setup(Path(device_config_file))

        config = uvicorn.Config(
            flowchem.http.app,
            host=host,
            port=flowchem.port,
            log_level="info",
            timeout_keep_alive=3600,
        )
        server = uvicorn.Server(config)
        logger.info("Click on http://127.0.0.1:8000 to access device server.")
        await server.serve()

    asyncio.run(main_loop())


if __name__ == "__main__":
    main()



"""Bronkhorst El-flow mass flow controller (MFC) device driver."""
import asyncio

# Manufacturer package, see https://bronkhorst-propar.readthedocs.io/en/latest/introduction.html.
import propar
from loguru import logger

from flowchem import ureg
from flowchem.devices.bronkhorst.el_flow_component import EPCComponent, MFCComponent
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.utils.people import wei_hsin


def isfloat(num):
    try:
        float(num)
        return True
    except ValueError:
        return False


class EPC(FlowchemDevice):
    """
    A class to represent an Electronic Pressure Controller (EPC) device from Bronkhorst.

    Attributes:
    -----------
    DEFAULT_CONFIG : dict
        Default configuration for the EPC device.
    port : str
        The port to which the EPC device is connected (Serial connection).
    channel : int
        The communication channel of the EPC device.
        Some instruments are a single node (one address) but contain multiple channels (for multiple sensors). To
        connect to a specific channel, specify the channel when creating an instrument instance. If no channel is
        specified the first channel will be used.
    address : int
        The address of the EPC device.
    max_pressure : float
        The maximum pressure of the EPC device in bar.
    id : str
        The identifier of the connected EPC device.

    Methods:
    --------
    initialize():
        Initialize the EPC device and set it to 0 bar.
    set_pressure(pressure: str):
        Set the pressure of the EPC device.
    get_pressure() -> float:
        Get the current pressure of the EPC device in bar.
    get_pressure_percentage() -> float:
        Get the current pressure of the EPC device as a percentage of the maximum pressure.
    """
    DEFAULT_CONFIG = {"channel": 1, "baudrate": 38400}  # "address": 0x80

    def __init__(
        self,
        port: str,
        name="",
        channel: int = 1,
        address: int = 0x80,
        max_pressure: float = 10,  # bar = 100 % = 32000
    ) -> None:
        """
        Constructs all the necessary attributes for the EPC object.

        Parameters:
        -----------
        port : str
            The port to which the EPC device is connected.
        name : str, optional
            The name of the EPC device (default is an empty string).
        channel : int, optional
            The communication channel of the EPC device (default is 1).        # Todo - explain better it
        address : int, optional
            The address of the EPC device (default is 0x80).                   # Todo - explain better it
        max_pressure : float, optional
            The maximum pressure of the EPC device in bar (default is 10).     # Todo - explain better it
        """
        self.port = port
        self.channel = channel
        self.address = address
        self.max_pressure = max_pressure
        super().__init__(name)

        # Metadata
        self.device_info.authors = [wei_hsin]
        self.device_info.manufacturer = "Bronkhorst"

        try:
            self.el_press = propar.instrument(
                self.port, address=self.address, channel=self.channel
            )
        except OSError as e:
            raise ConnectionError(f"Error connecting to {self.port} -- {e}") from e

        self.id = self.el_press.id
        logger.debug(f"Connected {self.id} to {self.port}")

    async def initialize(self):
        """Initialize the EPC device and set it to 0 bar."""
        await self.set_pressure("0 bar")
        self.components.append(EPCComponent("EPC", self))

    async def set_pressure(self, pressure: str):
        """
        Set the pressure setpoint of the EPC device.

        Parameters:
        -----------
        pressure : str
            The desired pressure to set. If no units are provided, bar is assumed.
        """
        if pressure.isnumeric() or isfloat(pressure):
            pressure = pressure + "bar"
            logger.warning("No units provided to set_pressure, assuming bar.")
        set_p = ureg.Quantity(pressure)
        set_n = round(set_p.m_as("bar") * 32000 / self.max_pressure)
        if set_n > 32000:
            self.el_press.setpoint = 32000
            logger.debug(
                "setting higher than maximum flow rate! set the flow rate to 100%",
            )
        else:
            self.el_press.setpoint = set_n
            logger.debug(f"set the pressure to {set_n / 320}%")

    async def get_pressure(self) -> float:
        """
        Get the current pressure of the EPC device in bar.

        Returns:
        --------
        float
            The current pressure in bar.
        """
        m_num = float(self.el_press.measure)
        return m_num / 32000 * self.max_pressure

    async def get_pressure_percentage(self) -> float:
        """
        Get the current pressure of the EPC device as a percentage of the maximum pressure.

        Returns:
        --------
        float
            The current pressure as a percentage of the maximum pressure.
        """
        m_num = float(self.el_press.measure)
        return m_num / 320


class MFC(FlowchemDevice):
    """
   A class to represent a Mass Flow Controller (MFC) device.

   Attributes:
   -----------
   DEFAULT_CONFIG : dict
       Default configuration for the MFC device.
   port : str
       The port to which the MFC device is connected.
   channel : int
       The communication channel of the MFC device.
   address : int
       The address of the MFC device.
   max_flow : float
       The maximum flow rate of the MFC device in ml/min.
   id : str
       The identifier of the connected MFC device.

   Methods:
   --------
   initialize():
       Ensure connection and initialize the MFC device.
   set_flow_setpoint(flowrate: str):
       Set the flow rate setpoint of the MFC device.
   get_flow_setpoint() -> float:
       Get the current flow rate of the MFC device in ml/min.
   get_flow_percentage() -> float:
       Get the current flow rate of the MFC device as a percentage of the maximum flow rate.
   """
    DEFAULT_CONFIG = {"channel": 1, "baudrate": 38400}  # "address": 0x80

    def __init__(
        self,
        port: str,
        name="",
        channel: int = 1,
        address: int = 0x80,
        max_flow: float = 9,  # ml / min = 100 % = 32000
    ) -> None:
        """
        Constructs all the necessary attributes for the MFC object.

        Parameters:
        -----------
        port : str
            The port to which the MFC device is connected.
        name : str, optional
            The name of the MFC device (default is an empty string).
        channel : int, optional
            The communication channel of the MFC device (default is 1).
        address : int, optional
            The address of the MFC device (default is 0x80).
        max_flow : float, optional
            The maximum flow rate of the MFC device in ml/min (default is 9).
        """
        self.port = port
        self.channel = channel
        self.address = address
        self.max_flow = max_flow
        super().__init__(name)
        self.max_flow = max_flow

        # Metadata
        self.device_info.model = "EL-FLOW"

        try:
            self.el_flow = propar.instrument(
                self.port, address=self.address, channel=self.channel
            )
        except OSError as e:
            raise ConnectionError(f"Error connecting to {self.port} -- {e}") from e
        self.id = self.el_flow.id
        logger.debug(f"Connected {self.id} to {self.port}")

    async def initialize(self):
        """Ensure connection and initialize the MFC device."""
        await self.set_flow_setpoint("0 ul/min")
        self.components.append(
            MFCComponent("MFC", self),
        )

    async def set_flow_setpoint(self, flowrate: str):
        """
        Set the flow rate setpoint of the MFC device in ml/min.

        Parameters:
        -----------
        flowrate : str
            The desired flow rate to set. If no units are provided, ml/min is assumed.
        """
        if flowrate.isnumeric() or isfloat(flowrate):
            flowrate = flowrate + "ml/min"
            logger.warning(
                "No units provided to set_flow_rate, assuming milliliter/minutes.",
            )

        set_f = ureg.Quantity(flowrate)
        set_n = round(set_f.m_as("ml/min") * 32000 / self.max_flow)
        if set_n > 32000:
            self.el_flow.setpoint = 32000
            logger.debug(
                "setting higher than maximum flow rate! set the flow rate to 100%",
            )
        else:
            self.el_flow.setpoint = set_n
            logger.debug(f"set the flow rate to {set_n / 320}%")

    async def get_flow_setpoint(self) -> float:
        """
        Get the current flow rate of the MFC device in ml/min.

        Returns:
        --------
        float
            The current flow rate in ml/min.
        """
        m_num = float(self.el_flow.measure)
        return m_num / 32000 * self.max_flow

    async def get_flow_percentage(self) -> float:
        """
        Get the current flow rate of the MFC device as a percentage of the maximum flow rate.

        Returns:
        --------
        float
            The current flow rate as a percentage of the maximum flow rate.
        """
        m_num = float(self.el_flow.measure)
        return m_num / 320


async def gas_flow(port: str, target_flowrate: str, reaction_time: float):
    Oxygen_flow = MFC(port, max_flow=10, address=6)
    await Oxygen_flow.initialize()
    await Oxygen_flow.set_flow_setpoint("900 ul/min")  # 10%
    # await asyncio.sleep(2)
    # for n in range(100):
    #     print(f"{n} time: {await Oxygen_flow.get_flow_setpoint()}%")
    #
    # # Oxygen_flow.set_flow_setpoint(target_point*32000/9.0)
    # # await asyncio.sleep(reaction_time*60)
    print(Oxygen_flow.id)
    await Oxygen_flow.set_flow_setpoint("0 ml/min")


async def mutiple_connect():
    flow = MFC("COM7", address=1, max_flow=10)
    EPC("COM7", address=2, max_pressure=10)
    MFC("COM7", address=6, max_flow=10)
    # O2_id = O2_flow.get_id
    # print(await pressure.get_id)
    # print(await flow.get_id)
    await flow.set_flow_setpoint("0.5")
    await flow.set_flow_setpoint("0")


def find_devices_info(port: str):
    """It is also possible to only create a master.
    This removes some abstraction offered by the instrument class,
    such as the setpoint and measure properties,
    the readParameter and writeParameter functions,
    and having to supply the node number on each read/write parameter call.
    """
    # Create the master
    master = propar.master(port, 38400)

    # Get nodes on the network
    nodes = master.get_nodes()

    # Read the usertag of all nodes
    for node in nodes:
        print(node)
        user_tag = master.read(node["address"], 113, 6, propar.PP_TYPE_STRING)
        print(user_tag)


if __name__ == "__main__":
    # find_devices_info("COM7")

    async def multiple_connect():
        flow = MFC("COM7", address=1, max_flow=10)
        pressure = EPC("COM7", address=2, max_pressure=10)
        O2_flow = MFC("COM7", address=6, max_flow=10)

        print(pressure.id)
        print(flow.id)
        print(O2_flow.id)

    asyncio.run(multiple_connect())

    db = propar.database()
    parameters = db.get_parameters([8, 9, 11, 142])
